import os

import numpy as np
import pyarrow as pa

import vaex


if vaex.utils.has_c_extension:
    from .superstrings import *

def array(ar):
    if isinstance(ar, (tuple, list)):
        ar = np.asarray(ar)
    if isinstance(ar, np.ndarray):
        return StringArray(ar.astype('O'))
    elif isinstance(ar, (StringList32, StringList64)):
        return ar
    else:
        raise ValueError('Cannot convert %r to a string array' % ar)

def to_string_sequence(strings):
    if not vaex.array_types.is_arrow_array(strings):
        strings = pa.array(strings)
    strings = vaex.column.ColumnStringArrow.from_arrow(strings)
    return strings.string_sequence


def to_arrow(ar):
    if isinstance(ar, (StringList32, StringList64)):
        col = vaex.column.ColumnStringArrow.from_string_sequence(ar)
        ar = pa.array(col)
    elif isinstance(ar, vaex.array_types.supported_arrow_array_types):
        pass
    else:
        raise TypeError(f'{ar} cannot be converted to array string array')
    return ar



import ast
import copy
import os
import base64
import datetime
from pydoc import doc
import time
import cloudpickle as pickle
import functools
import operator
import six
import collections
import weakref

from future.utils import with_metaclass
import numpy as np
import pandas as pd
import tabulate
import pyarrow as pa

import vaex.cache
import vaex.hash
import vaex.tasks
import vaex.array_types
import vaex.serialize
from vaex.utils import _ensure_strings_from_expressions, _ensure_string_from_expression
from vaex.column import ColumnString, _to_string_sequence
from .hash import counter_type_from_dtype
from vaex.datatype import DataType
from vaex.docstrings import docsubst
from . import expresso


try:
    from StringIO import StringIO
except ImportError:
    from io import BytesIO as StringIO

try:
    collectionsAbc = collections.abc
except AttributeError:
    collectionsAbc = collections


# TODO: repeated from dataframe.py
default_shape = 128
PRINT_MAX_COUNT = 10

expression_namespace = {}
expression_namespace['nan'] = np.nan


expression_namespace = {}
expression_namespace['nan'] = np.nan


_binary_ops = [
    dict(code="+", name='add', op=operator.add),
    dict(code="in", name='contains', op=operator.contains),
    dict(code="/", name='truediv', op=operator.truediv),
    dict(code="//", name='floordiv', op=operator.floordiv),
    dict(code="&", name='and', op=operator.and_),
    dict(code="^", name='xor', op=operator.xor),

    dict(code="|", name='or', op=operator.or_),
    dict(code="**", name='pow', op=operator.pow),
    dict(code="is", name='is', op=operator.is_),
    dict(code="is not", name='is_not', op=operator.is_not),

    dict(code="<<", name='lshift', op=operator.lshift),
    dict(code="%", name='mod', op=operator.mod),
    dict(code="*", name='mul', op=operator.mul),

    dict(code=">>", name='rshift', op=operator.rshift),
    dict(code="-", name='sub', op=operator.sub),

    dict(code="<", name='lt', op=operator.lt),
    dict(code="<=", name='le', op=operator.le),
    dict(code="==", name='eq', op=operator.eq),
    dict(code="!=", name='ne', op=operator.ne),
    dict(code=">=", name='ge', op=operator.ge),
    dict(code=">", name='gt', op=operator.gt),
]
if hasattr(operator, 'div'):
    _binary_ops.append(dict(code="/", name='div', op=operator.div))
if hasattr(operator, 'matmul'):
    _binary_ops.append(dict(code="@", name='matmul', op=operator.matmul))

reversable = 'add sub mul matmul truediv floordiv mod divmod pow lshift rshift and xor or'.split()

_unary_ops = [
    dict(code="~", name='invert', op=operator.invert),
    dict(code="-", name='neg', op=operator.neg),
    dict(code="+", name='pos', op=operator.pos),
]


class Meta(type):
    def __new__(upperattr_metaclass, future_class_name,
                future_class_parents, attrs):
        # attrs = {}
        for op in _binary_ops:
            def wrap(op=op):
                def f(a, b):
                    if isinstance(a, pa.Scalar):
                        a = a.as_py()
                    if isinstance(b, pa.Scalar):
                        b = b.as_py()
                    self = a
                    # print(op, a, b)
                    if isinstance(b, str) and self.dtype.is_datetime:
                        b = np.datetime64(b)
                    if self.df.is_category(self.expression) and self.df._future_behaviour and not isinstance(b, Expression):
                        labels = self.df.category_labels(self.expression)
                        if b not in labels:
                            raise ValueError(f'Value {b} not present in {labels}')
                        b = labels.index(b)
                        a = self.index_values()
                    try:
                        stringy = isinstance(b, str) or b.is_string()
                    except:
                        # this can happen when expression is a literal, like '1' (used in propagate_unc)
                        # which causes the dtype to fail
                        stringy = False
                    if stringy:
                        if isinstance(b, str):
                            b = repr(b)
                        if op['code'] == '==':
                            expression = 'str_equals({0}, {1})'.format(a.expression, b)
                        elif op['code'] == '!=':
                            expression = 'str_notequals({0}, {1})'.format(a.expression, b)
                        elif op['code'] == '+':
                            expression = 'str_cat({0}, {1})'.format(a.expression, b)
                        else:
                            raise ValueError('operand %r not supported for string comparison' % op['code'])
                        return Expression(self.ds, expression=expression)
                    else:
                        if isinstance(b, Expression):
                            assert b.ds == a.ds
                            b = b.expression
                        elif isinstance(b, (np.timedelta64)):
                            unit, step = np.datetime_data(b.dtype)
                            assert step == 1
                            b = b.astype(np.uint64).item()
                            b = f'scalar_timedelta({b}, {unit!r})'
                        elif isinstance(b, (np.datetime64)):
                            b = f'scalar_datetime("{b}")'
                        elif isinstance(b, np.ndarray) and b.ndim == 0 and vaex.dtype_of(b).is_datetime:
                            b = f'scalar_datetime("{b}")'
                        elif isinstance(b, np.ndarray) and b.ndim == 0 and vaex.dtype_of(b).is_timedelta:
                            unit, step = np.datetime_data(b.dtype)
                            assert step == 1
                            b = b.astype(np.uint64).item()
                            b = f'scalar_timedelta({b}, {unit!r})'
                        expression = '({0} {1} {2})'.format(a.expression, op['code'], b)
                    return Expression(self.ds, expression=expression)
                attrs['__%s__' % op['name']] = f
                if op['name'] in reversable:
                    def f(a, b):
                        if isinstance(a, pa.Scalar):
                            a = a.as_py()
                        if isinstance(b, pa.Scalar):
                            b = b.as_py()
                        self = a
                        if isinstance(b, str):
                            if op['code'] == '+':
                                expression = 'str_cat({1}, {0})'.format(a.expression, repr(b))
                            else:
                                raise ValueError('operand %r not supported for string comparison' % op['code'])
                            return Expression(self.ds, expression=expression)
                        else:
                            if isinstance(b, Expression):
                                assert b.ds == a.ds
                                b = b.expression
                            expression = '({2} {1} {0})'.format(a.expression, op['code'], b)
                            return Expression(self.ds, expression=expression)
                    attrs['__r%s__' % op['name']] = f

            wrap(op)
        for op in _unary_ops:
            def wrap(op=op):
                def f(a):
                    if isinstance(a, pa.Scalar):
                        a = a.as_py()
                    self = a
                    expression = '{0}({1})'.format(op['code'], a.expression)
                    return Expression(self.ds, expression=expression)
                attrs['__%s__' % op['name']] = f
            wrap(op)
        return type(future_class_name, future_class_parents, attrs)


class DateTime(object):
    """DateTime operations

    Usually accessed using e.g. `df.birthday.dt.dayofweek`
    """
    def __init__(self, expression):
        self.expression = expression


class TimeDelta(object):
    """TimeDelta operations

    Usually accessed using e.g. `df.delay.td.days`
    """
    def __init__(self, expression):
        self.expression = expression


class StringOperations(object):
    """String operations.

    Usually accessed using e.g. `df.name.str.lower()`
    """
    def __init__(self, expression):
        self.expression = expression


class StringOperationsPandas(object):
    """String operations using Pandas Series (much slower)"""
    def __init__(self, expression):
        self.expression = expression


class StructOperations(collections.abc.Mapping):
    """Struct Array operations.

    Usually accessed using e.g. `df.name.struct.get('field1')`

    """

    def __init__(self, expression):
        self.expression = expression
        self._array = self.expression.values

    def __iter__(self):
        for name in self.keys():
            yield name

    def __getitem__(self, key):
        """Return struct field by either field name (string) or index position (index).

        In case of ambiguous field names, a `LookupError` is raised.

        """

        self._assert_struct_dtype()
        return self.get(key)

    def __len__(self):
        """Return the number of struct fields contained in struct array.

        """

        self._assert_struct_dtype()
        return len(self._array.type)

    def keys(self):
        """Return all field names contained in struct array.

        :returns: list of field names.

        Example:

        >>> import vaex
        >>> import pyarrow as pa
        >>> array = pa.StructArray.from_arrays(arrays=[[1,2], ["a", "b"]], names=["col1", "col2"])
        >>> df = vaex.from_arrays(array=array)
        >>> df
        # 	array
        0	{'col1': 1, 'col2': 'a'}
        1	{'col1': 2, 'col2': 'b'}

        >>> df.array.struct.keys()
        ["col1", "col2"]

        """

        self._assert_struct_dtype()
        return [field.name for field in self._array.type]

    def values(self):
        """Return all fields as vaex expressions.

        :returns: list of vaex expressions corresponding to each field in struct.

        Example:

        >>> import vaex
        >>> import pyarrow as pa
        >>> array = pa.StructArray.from_arrays(arrays=[[1,2], ["a", "b"]], names=["col1", "col2"])
        >>> df = vaex.from_arrays(array=array)
        >>> df
        # 	array
        0	{'col1': 1, 'col2': 'a'}
        1	{'col1': 2, 'col2': 'b'}

        >>> df.array.struct.values()
        [Expression = struct_get(array, 0)
         Length: 2 dtype: int64 (expression)
         -----------------------------------
         0  1
         1  2,
         Expression = struct_get(array, 1)
         Length: 2 dtype: string (expression)
         ------------------------------------
         0  a
         1  b]

        """

        self._assert_struct_dtype()
        return [self[i] for i in range(len(self))]

    def items(self):
        """Return all fields with names along with corresponding vaex expressions.

        :returns: list of tuples with field names and fields as vaex expressions.

        Example:

        >>> import vaex
        >>> import pyarrow as pa
        >>> array = pa.StructArray.from_arrays(arrays=[[1,2], ["a", "b"]], names=["col1", "col2"])
        >>> df = vaex.from_arrays(array=array)
        >>> df
        # 	array
        0	{'col1': 1, 'col2': 'a'}
        1	{'col1': 2, 'col2': 'b'}

        >>> df.array.struct.items()
        [('col1',
          Expression = struct_get(array, 0)
          Length: 2 dtype: int64 (expression)
          -----------------------------------
          0  1
          1  2),
         ('col2',
          Expression = struct_get(array, 1)
          Length: 2 dtype: string (expression)
          ------------------------------------
          0  a
          1  b)]

        """

        self._assert_struct_dtype()
        return list(zip(self.keys(), self.values()))

    @property
    def dtypes(self):
        """Return all field names along with corresponding types.

        :returns: a pandas series with keys as index and types as values.

        Example:

        >>> import vaex
        >>> import pyarrow as pa
        >>> array = pa.StructArray.from_arrays(arrays=[[1,2], ["a", "b"]], names=["col1", "col2"])
        >>> df = vaex.from_arrays(array=array)
        >>> df
        # 	array
        0	{'col1': 1, 'col2': 'a'}
        1	{'col1': 2, 'col2': 'b'}

        >>> df.array.struct.dtypes
        col1     int64
        col2    string
        dtype: object

        """

        self._assert_struct_dtype()
        dtypes = (field.type for field in self._array.type)
        vaex_dtypes = [DataType(x) for x in dtypes]

        return pd.Series(vaex_dtypes, index=self.keys())

    def _assert_struct_dtype(self):
        """Ensure that struct operations are only called on valid struct dtype.

        """

        from vaex.struct import assert_struct_dtype
        assert_struct_dtype(self._array)


class Expression(with_metaclass(Meta)):
    """Expression class"""
    def __init__(self, ds, expression, ast=None, _selection=False):
        import vaex.dataframe
        self.ds : vaex.dataframe.DataFrame = ds
        assert not isinstance(ds, Expression)
        if isinstance(expression, Expression):
            expression = expression.expression
        if expression is None and ast is None:
            raise ValueError('Not both expression and the ast can be None')
        self._ast = ast
        self._expression = expression
        self.df._expressions.append(weakref.ref(self))
        self._ast_names = None
        self._selection = _selection  # selection have an extra scope

    @property
    def _label(self):
        '''If a column is an invalid identified, the expression is df['long name']
        This will return 'long name' in that case, otherwise simply the expression
        '''
        ast = self.ast
        if isinstance(ast, expresso._ast.Subscript):
            value = ast.slice.value
            if isinstance(value, expresso.ast_Str):
                return value.s
            if isinstance(value, str):  # py39+
                return value
        return self.expression

    def fingerprint(self):
        fp = vaex.cache.fingerprint(self.expression, self.df.fingerprint(dependencies=self.dependencies()))
        return f'expression-{fp}'

    def copy(self, df=None):
        """Efficiently copies an expression.

        Expression objects have both a string and AST representation. Creating
        the AST representation involves parsing the expression, which is expensive.

        Using copy will deepcopy the AST when the expression was already parsed.

        :param df: DataFrame for which the expression will be evaluated (self.df if None)
        """
        # expression either has _expression or _ast not None
        if df is None:
            df = self.df
        if self._expression is not None:
            expression = Expression(df, self._expression)
            if self._ast is not None:
                expression._ast = copy.deepcopy(self._ast)
        elif self._ast is not None:
            expression = Expression(df, copy.deepcopy(self._ast))
            if self._ast is not None:
                expression._ast = self._ast
        return expression

    @property
    def ast(self):
        """Returns the abstract syntax tree (AST) of the expression"""
        if self._ast is None:
            self._ast = expresso.parse_expression(self.expression)
        return self._ast

    @property
    def ast_names(self):
        if self._ast_names is None:
            self._ast_names = expresso.names(self.ast)
        return self._ast_names

    @property
    def _ast_slices(self):
        return expresso.slices(self.ast)

    @property
    def expression(self):
        if self._expression is None:
            self._expression = expresso.node_to_string(self.ast)
        return self._expression

    @expression.setter
    def expression(self, value):
        # if we reassign to expression, we clear the ast cache
        if value != self._expression:
            self._expression = value
            self._ast = None

    def __bool__(self):
        """Cast expression to boolean. Only supports (<expr1> == <expr2> and <expr1> != <expr2>)

        The main use case for this is to support assigning to traitlets. e.g.:

        >>> bool(expr1 == expr2)

        This will return True when expr1 and expr2 are exactly the same (in string representation). And similarly for:

        >>> bool(expr != expr2)

        All other cases will return True.
        """
        # this is to make traitlets detect changes
        import _ast
        if isinstance(self.ast, _ast.Compare) and len(self.ast.ops) == 1 and isinstance(self.ast.ops[0], _ast.Eq):
            return expresso.node_to_string(self.ast.left) == expresso.node_to_string(self.ast.comparators[0])
        if isinstance(self.ast, _ast.Compare) and len(self.ast.ops) == 1 and isinstance(self.ast.ops[0], _ast.NotEq):
            return expresso.node_to_string(self.ast.left) != expresso.node_to_string(self.ast.comparators[0])
        return True


    @property
    def df(self):
        # lets gradually move to using .df
        return self.ds

    @property
    def dtype(self):
        return self.df.data_type(self)

    # TODO: remove this method?
    def data_type(self, array_type=None, axis=0) -> DataType:
        return self.df.data_type(self, axis=axis)

    @property
    def shape(self):
        return self.df._shape_of(self)

    @property
    def ndim(self):
        return 1 if self.dtype.is_list else len(self.df._shape_of(self))

    def to_arrow(self, convert_to_native=False):
        '''Convert to Apache Arrow array (will byteswap/copy if convert_to_native=True).'''
        values = self.values
        return vaex.array_types.to_arrow(values, convert_to_native=convert_to_native)

    def __arrow_array__(self, type=None):
        values = self.to_arrow()
        return pa.array(values, type=type)

    def to_numpy(self, strict=True):
        """Return a numpy representation of the data"""
        values = self.values
        return vaex.array_types.to_numpy(values, strict=strict)

    def to_dask_array(self, chunks="auto"):
        import dask.array as da
        import uuid
        dtype = self.dtype
        chunks = da.core.normalize_chunks(chunks, shape=self.shape, dtype=dtype.numpy)
        name = 'vaex-expression-%s' % str(uuid.uuid1())
        def getitem(df, item):
            assert len(item) == 1
            item = item[0]
            start, stop, step = item.start, item.stop, item.step
            assert step in [None, 1]
            return self.evaluate(start, stop, parallel=False)
        if hasattr(da.core, "getem"):
            dsk = da.core.getem(name, chunks, getitem=getitem, shape=self.shape, dtype=dtype.numpy)
            dsk[name] = self
            return da.Array(dsk, name, chunks, dtype=dtype.numpy)
        else:
            dsk = da.core.graph_from_arraylike(self, name=name, chunks=chunks, getitem=getitem, shape=self.shape, dtype=dtype.numpy)
            return da.Array(dsk, name, chunks, dtype=dtype.numpy)

    def to_pandas_series(self):
        """Return a pandas.Series representation of the expression.

        Note: Pandas is likely to make a memory copy of the data.
        """
        import pandas as pd
        return pd.Series(self.values)

    def __getitem__(self, slicer):
        """Provides row and optional field access (struct arrays) via bracket notation.

        Examples:

        >>> import vaex
        >>> import pyarrow as pa
        >>> array = pa.StructArray.from_arrays(arrays=[[1, 2, 3], ["a", "b", "c"]], names=["col1", "col2"])
        >>> df = vaex.from_arrays(array=array, integer=[5, 6, 7])
        >>> df
        # 	array 	                    integer
        0	{'col1': 1, 'col2': 'a'}	5
        1	{'col1': 2, 'col2': 'b'}	6
        2	{'col1': 3, 'col2': 'c'}	7

        >>> df.integer[1:]
        Expression = integer
        Length: 2 dtype: int64 (column)
        -------------------------------
        0  6
        1  7

        >>> df.array[1:]
        Expression = array
        Length: 2 dtype: struct<col1: int64, col2: string> (column)
        -----------------------------------------------------------
        0  {'col1': 2, 'col2': 'b'}
        1  {'col1': 3, 'col2': 'c'}

        >>> df.array[:, "col1"]
        Expression = struct_get(array, 'col1')
        Length: 3 dtype: int64 (expression)
        -----------------------------------
        0  1
        1  2
        2  3

        >>> df.array[1:, ["col1"]]
        Expression = struct_project(array, ['col1'])
        Length: 2 dtype: struct<col1: int64> (expression)
        -------------------------------------------------
        0  {'col1': 2}
        1  {'col1': 3}

        >>> df.array[1:, ["col2", "col1"]]
        Expression = struct_project(array, ['col2', 'col1'])
        Length: 2 dtype: struct<col2: string, col1: int64> (expression)
        ---------------------------------------------------------------
        0  {'col2': 'b', 'col1': 2}
        1  {'col2': 'c', 'col1': 3}

        """

        if isinstance(slicer, slice):
            indices = slicer
            fields = None
        elif isinstance(slicer, tuple) and len(slicer) == 2:
            indices, fields = slicer
        else:
            raise NotImplementedError

        if indices != slice(None):
            expr = self.df[indices][self.expression]
        else:
            expr = self

        if fields is None:
            return expr
        elif isinstance(fields, (int, str)):
            if self.dtype.is_struct:
                return expr.struct.get(fields)
            elif self.ndim == 2:
                if not isinstance(fields, int):
                    raise TypeError(f'Expected an integer, not {type(fields)}')
                else:
                    return expr.getitem(fields)
            else:
                raise TypeError(f'Only getting struct fields or 2d columns supported')
        elif isinstance(fields, (tuple, list)):
            return expr.struct.project(fields)
        else:
            raise TypeError("Invalid type provided. Needs to be None, str or list/tuple.")

    def __abs__(self):
        """Returns the absolute value of the expression"""
        return self.abs()

    @property
    def dt(self):
        """Gives access to datetime operations via :py:class:`DateTime`"""
        return DateTime(self)

    @property
    def td(self):
        """Gives access to timedelta operations via :py:class:`TimeDelta`"""
        return TimeDelta(self)

    @property
    def str(self):
        """Gives access to string operations via :py:class:`StringOperations`"""
        return StringOperations(self)

    @property
    def str_pandas(self):
        """Gives access to string operations via :py:class:`StringOperationsPandas` (using Pandas Series)"""
        return StringOperationsPandas(self)

    @property
    def struct(self):
        """Gives access to struct operations via :py:class:`StructOperations`"""
        return StructOperations(self)

    @property
    def values(self):
        return self.evaluate()

    def derivative(self, var, simplify=True):
        var = _ensure_string_from_expression(var)
        return self.__class__(self.ds, expresso.derivative(self.ast, var, simplify=simplify))

    def expand(self, stop=[]):
        """Expand the expression such that no virtual columns occurs, only normal columns.

        Example:

        >>> df = vaex.example()
        >>> r = np.sqrt(df.data.x**2 + df.data.y**2)
        >>> r.expand().expression
        'sqrt(((x ** 2) + (y ** 2)))'

        """
        stop = _ensure_strings_from_expressions(stop)
        def translate(id):
            if id in self.ds.virtual_columns and id not in stop:
                return self.ds.virtual_columns[id]
        expr = expresso.translate(self.ast, translate)
        return Expression(self.ds, expr)

    def dependencies(self):
        '''Get all dependencies of this expression, including ourselves'''
        return self.variables(ourself=True)

    def variables(self, ourself=False, expand_virtual=True, include_virtual=True):
        """Return a set of variables this expression depends on.

        Example:

        >>> df = vaex.example()
        >>> r = np.sqrt(df.data.x**2 + df.data.y**2)
        >>> r.variables()
        {'x', 'y'}
        """
        variables = set()
        def record(varname):
            # always do this for selection
            if self._selection and self.df.has_selection(varname):
                selection = self.df.get_selection(varname)
                variables.update(selection.dependencies(self.df))
            # do this recursively for virtual columns
            if varname in self.ds.virtual_columns and varname not in variables:
                if (include_virtual and (varname != self.expression)) or (varname == self.expression and ourself):
                    variables.add(varname)
                if expand_virtual:
                    variables.update(self.df[self.df.virtual_columns[varname]].variables(ourself=include_virtual, include_virtual=include_virtual))
            # we usually don't want to record ourself
            elif varname != self.expression or ourself:
                variables.add(varname)

        expresso.translate(self.ast, record)
        # df is a buildin, don't record it, if df is a column name, it will be collected as
        # df['df']
        variables -= {'df'}
        for varname in self._ast_slices:
            if varname in self.df.virtual_columns and varname not in variables:
                if (include_virtual and (f"df['{varname}']" != self.expression)) or (f"df['{varname}']" == self.expression and ourself):
                    variables.add(varname)
                if expand_virtual:
                    if varname in self.df.virtual_columns:
                        variables |= self.df[self.df.virtual_columns[varname]].variables(ourself=include_virtual, include_virtual=include_virtual)
            elif f"df['{varname}']" != self.expression or ourself:
                variables.add(varname)

        return variables

    def _graph(self):
        """"Return a graph containing the dependencies of this expression
        Structure is:
            [<string expression>, <function name if callable>, <function object if callable>, [subgraph/dependencies, ....]]
        """
        expression = self.expression

        def walk(node):
            if isinstance(node, six.string_types):
                if node in self.ds.virtual_columns:
                    ex = Expression(self.ds, self.ds.virtual_columns[node])
                    return [node, None, None, [ex._graph()]]
                else:
                    return node
            else:
                fname, node_repr, deps = node
                if len(node_repr) > 30:  # clip too long expressions
                    node_repr = node_repr[:26] + ' ....'
                deps = [walk(dep) for dep in deps]
                obj = self.ds.functions.get(fname)
                # we don't want the wrapper, we want the underlying object
                if isinstance(obj, Function):
                    obj = obj.f
                if isinstance(obj, FunctionSerializablePickle):
                    obj = obj.f
                return [node_repr, fname, obj, deps]
        return walk(expresso._graph(expression))

    def _graphviz(self, dot=None):
        """Return a graphviz.Digraph object with a graph of the expression"""
        from graphviz import Graph, Digraph
        node = self._graph()
        dot = dot or Digraph(comment=self.expression)
        def walk(node):
            if isinstance(node, six.string_types):
                dot.node(node, node)
                return node, node
            else:
                node_repr, fname, fobj, deps = node
                node_id = node_repr
                dot.node(node_id, node_repr)
                for dep in deps:
                    dep_id, dep = walk(dep)
                    dot.edge(node_id, dep_id)
                return node_id, node
        walk(node)
        return dot

    def __str__(self):
        return self.expression

    # def __array__(self, dtype=None):
    #     '''For casting to a numpy array

    #     Example:
    #         >>> np.array(ds.x**2)

    #     '''
    #     return self.ds.evaluate(self)

    def tolist(self, i1=None, i2=None):
        '''Short for expr.evaluate().tolist()'''
        values = self.evaluate(i1=i1, i2=i2)
        if isinstance(values, (pa.Array, pa.ChunkedArray)):
            return values.to_pylist()
        return values.tolist()

    if not os.environ.get('VAEX_DEBUG', ''):
        def __repr__(self):
            return self._repr_plain_()

    def _repr_plain_(self):
        from .formatting import _format_value
        def format(values):
            for i in range(len(values)):
                value = values[i]
                yield _format_value(value)
        colalign = ("right",) * 2
        try:
            N = len(self.ds)
            if N <= PRINT_MAX_COUNT:
                values = format(self.evaluate(0, N))
                values = tabulate.tabulate([[i, k] for i, k in enumerate(values)], tablefmt='plain', colalign=colalign)
            else:
                values_head = format(self.evaluate(0, PRINT_MAX_COUNT//2))
                values_tail = format(self.evaluate(N - PRINT_MAX_COUNT//2, N))
                values_head = list(zip(range(PRINT_MAX_COUNT//2), values_head)) +\
                              list(zip(range(N - PRINT_MAX_COUNT//2, N), values_tail))
                values = tabulate.tabulate([k for k in values_head], tablefmt='plain', colalign=colalign)
                values = values.split('\n')
                width = max(map(len, values))
                separator = '\n' + '...'.center(width, ' ') + '\n'
                values = "\n".join(values[:PRINT_MAX_COUNT//2]) + separator + "\n".join(values[PRINT_MAX_COUNT//2:]) + '\n'
        except Exception as e:
            values = 'Error evaluating: %r' % e
        expression = self.expression
        if len(expression) > 60:
            expression = expression[:57] + '...'
        info = 'Expression = ' + expression + '\n'
        dtype = self.dtype
        if self.expression in self.ds.get_column_names(hidden=True):
            state = "column"
        elif self.expression in self.ds.get_column_names(hidden=True):
            state = "virtual column"
        else:
            state = "expression"
        line = 'Length: {:,} dtype: {} ({})\n'.format(len(self.ds), dtype, state)
        info += line
        info += '-' * (len(line)-1) + '\n'
        info += values
        return info

    def count(self, binby=[], limits=None, shape=default_shape, selection=False, delay=False, edges=False, progress=None):
        '''Shortcut for ds.count(expression, ...), see `Dataset.count`'''
        kwargs = dict(locals())
        del kwargs['self']
        kwargs['expression'] = self.expression
        return self.ds.count(**kwargs)

    def sum(self, axis=None, binby=[], limits=None, shape=default_shape, selection=False, delay=False, progress=None):
        '''Sum elements over given axis.

        If no axis is given, it will sum over all axes.

        For non list elements, this is a shortcut for ds.sum(expression, ...), see `Dataset.sum`.

        >>> list_data = [1, 2, None], None, [], [1, 3, 4, 5]
        >>> df = vaex.from_arrays(some_list=pa.array(list_data))
        >>> df.some_list.sum().item()  # will sum over all axis
        16
        >>> df.some_list.sum(axis=1).tolist()  # sums the list elements
        [3, None, 0, 13]

        :param int axis: Axis over which to determine the unique elements (None will flatten arrays or lists)
        '''
        expression = self
        if axis is None:
            dtype = self.dtype
            if dtype.is_list:
                axis = [0]
                while dtype.is_list:
                    axis.append(axis[-1] + 1)
                    dtype = dtype.value_type
            elif self.ndim > 1:
                axis = list(range(self.ndim))
            else:
                axis = [0]
        elif not isinstance(axis, list):
            axis = [axis]
            axis = list(set(axis))  # remove repeated elements
        dtype = self.dtype
        if self.ndim > 1:
            array_axes = axis.copy()
            if 0 in array_axes:
                array_axes.remove(0)
            expression = expression.array_sum(axis=array_axes)
            for i in array_axes:
                axis.remove(i)
                del i
            del array_axes
        elif 1 in axis:
            if self.dtype.is_list:
                expression = expression.list_sum()
                if axis:
                    axis.remove(1)
            else:
                raise ValueError(f'axis=1 not supported for dtype={dtype}')
        if axis and axis[0] != 0:
            raise ValueError(f'Only axis 0 or 1 is supported')
        if expression.ndim > 1:
            raise ValueError(f'Cannot sum non-scalar (ndim={expression.ndim})')
        if axis is None or 0 in axis:
            kwargs = dict(locals())
            del kwargs['self']
            del kwargs['axis']
            del kwargs['dtype']
            kwargs['expression'] = expression.expression
            return self.ds.sum(**kwargs)
        else:
            return expression

    def mean(self, binby=[], limits=None, shape=default_shape, selection=False, delay=False, progress=None):
        '''Shortcut for ds.mean(expression, ...), see `Dataset.mean`'''
        kwargs = dict(locals())
        del kwargs['self']
        kwargs['expression'] = self.expression
        return self.ds.mean(**kwargs)

    def std(self, binby=[], limits=None, shape=default_shape, selection=False, delay=False, progress=None):
        '''Shortcut for ds.std(expression, ...), see `Dataset.std`'''
        kwargs = dict(locals())
        del kwargs['self']
        kwargs['expression'] = self.expression
        return self.ds.std(**kwargs)

    def var(self, binby=[], limits=None, shape=default_shape, selection=False, delay=False, progress=None):
        '''Shortcut for ds.std(expression, ...), see `Dataset.var`'''
        kwargs = dict(locals())
        del kwargs['self']
        kwargs['expression'] = self.expression
        return self.ds.var(**kwargs)

    def skew(self, binby=[], limits=None, shape=default_shape, selection=False, delay=False, progress=None):
        '''Shortcut for df.skew(expression, ...), see `DataFrame.skew`'''
        kwargs = dict(locals())
        del kwargs['self']
        kwargs['expression'] = self.expression
        return self.df.skew(**kwargs)

    def kurtosis(self, binby=[], limits=None, shape=default_shape, selection=False, delay=False, progress=None):
        '''Shortcut for df.kurtosis(expression, ...), see `DataFrame.kurtosis`'''
        kwargs = dict(locals())
        del kwargs['self']
        kwargs['expression'] = self.expression
        return self.df.kurtosis(**kwargs)

    def minmax(self, binby=[], limits=None, shape=default_shape, selection=False, delay=False, progress=None):
        '''Shortcut for ds.minmax(expression, ...), see `Dataset.minmax`'''
        kwargs = dict(locals())
        del kwargs['self']
        kwargs['expression'] = self.expression
        return self.ds.minmax(**kwargs)

    def min(self, binby=[], limits=None, shape=default_shape, selection=False, delay=False, progress=None):
        '''Shortcut for ds.min(expression, ...), see `Dataset.min`'''
        kwargs = dict(locals())
        del kwargs['self']
        kwargs['expression'] = self.expression
        return self.ds.min(**kwargs)

    def max(self, binby=[], limits=None, shape=default_shape, selection=False, delay=False, progress=None):
        '''Shortcut for ds.max(expression, ...), see `Dataset.max`'''
        kwargs = dict(locals())
        del kwargs['self']
        kwargs['expression'] = self.expression
        return self.ds.max(**kwargs)

    def nop(self):
        """Evaluates expression, and drop the result, usefull for benchmarking, since vaex is usually lazy"""
        return self.ds.nop(self.expression)

    @property
    def transient(self):
        """If this expression is not transient (e.g. on disk) optimizations can be made"""
        return self.expand().expression not in self.ds.columns

    @property
    def masked(self):
        """Alias to df.is_masked(expression)"""
        return self.ds.is_masked(self.expression)

    @docsubst
    def value_counts(self, dropna=False, dropnan=False, dropmissing=False, ascending=False, progress=False, axis=None, delay=False):
        """Computes counts of unique values.

         WARNING:
          * If the expression/column is not categorical, it will be converted on the fly
          * dropna is False by default, it is True by default in pandas

        :param dropna: {dropna}
        :param dropnan: {dropnan}
        :param dropmissing: {dropmissing}
        :param ascending: when False (default) it will report the most frequent occuring item first
        :param progress: {progress}
        :param bool axis: Axis over which to determine the unique elements (None will flatten arrays or lists)
        :param bool delay: {delay}
        :returns: Pandas series containing the counts
        """
        from pandas import Series
        if axis is not None:
            raise ValueError('only axis=None is supported')
        if dropna:
            dropnan = True
            dropmissing = True

        progressbar = vaex.utils.progressbars(progress)
        task = vaex.tasks.TaskValueCounts(self.df, self.expression, dropnan=dropnan, dropmissing=dropmissing, ascending=ascending, axis=axis)
        self.df.executor.schedule(task)
        progressbar.add_task(task, "value counts")
        return self.df._delay(delay, task)


    @docsubst
    def unique(self, dropna=False, dropnan=False, dropmissing=False, selection=None, axis=None, limit=None, limit_raise=True, array_type='list', progress=None, delay=False):
        """Returns all unique values.

        :param dropna: {dropna}
        :param dropnan: {dropnan}
        :param dropmissing: {dropmissing}
        :param selection: {selection}
        :param bool axis: Axis over which to determine the unique elements (None will flatten arrays or lists)
        :param int limit: {limit}
        :param bool limit_raise: {limit_raise}
        :param bool array_type: {array_type}
        :param progress: {progress}
        :param bool delay: {delay}
        """
        return self.ds.unique(self, dropna=dropna, dropnan=dropnan, dropmissing=dropmissing, selection=selection, array_type=array_type, axis=axis, limit=limit, limit_raise=limit_raise, progress=progress, delay=delay)

    @docsubst
    def nunique(self, dropna=False, dropnan=False, dropmissing=False, selection=None, axis=None, limit=None, limit_raise=True, progress=None, delay=False):
        """Counts number of unique values, i.e. `len(df.x.unique()) == df.x.nunique()`.

        :param dropna: {dropna}
        :param dropnan: {dropnan}
        :param dropmissing: {dropmissing}
        :param selection: {selection}
        :param bool axis: Axis over which to determine the unique elements (None will flatten arrays or lists)
        :param int limit: {limit}
        :param bool limit_raise: {limit_raise}
        :param progress: {progress}
        :param bool delay: {delay}
        """
        def key_function():
            fp = vaex.cache.fingerprint(self.fingerprint(), dropna, dropnan, dropmissing, selection, axis, limit)
            return f'nunique-{fp}'
        @vaex.cache._memoize(key_function=key_function, delay=delay)
        def f():
            value = self.unique(dropna=dropna, dropnan=dropnan, dropmissing=dropmissing, selection=selection, axis=axis, limit=limit, limit_raise=limit_raise, array_type=None, progress=progress, delay=delay)
            if delay:
                return value.then(len)
            else:
                return len(value)
        return f()

    def countna(self):
        """Returns the number of Not Availiable (N/A) values in the expression.
        This includes missing values and np.nan values.
        """
        return self.isna().sum().item()  # so the output is int, not array

    def countnan(self):
        """Returns the number of NaN values in the expression."""
        return self.isnan().sum().item()  # so the output is int, not array

    def countmissing(self):
        """Returns the number of missing values in the expression."""
        return self.ismissing().sum().item()  # so the output is int, not array

    def evaluate(self, i1=None, i2=None, out=None, selection=None, parallel=True, array_type=None):
        return self.ds.evaluate(self, i1, i2, out=out, selection=selection, array_type=array_type, parallel=parallel)

    # TODO: it is not so elegant we need to have a custom version of this
    # it now also misses the docstring, reconsider how the the meta class auto
    # adds this method
    def fillna(self, value, fill_nan=True, fill_masked=True):
        expression = self._upcast_for(value)
        return self.ds.func.fillna(expression, value=value, fill_nan=fill_nan, fill_masked=fill_masked)

    def _upcast_for(self, value):
        # make sure the dtype is compatible with value
        expression = self
        dtype = self.dtype

        if dtype == int:
            required_dtype = vaex.utils.required_dtype_for_int(value, signed=dtype.is_signed)
            if required_dtype.itemsize > dtype.numpy.itemsize:
                expression = self.astype(str(required_dtype))
        return expression

    def fillmissing(self, value):
        '''Returns an array where missing values are replaced by value.

        See :`ismissing` for the definition of missing values.
        '''
        expression = self._upcast_for(value)
        return self.df.func.fillmissing(expression, value=value)

    def clip(self, lower=None, upper=None):
        return self.ds.func.clip(self, lower, upper)

    def jit_metal(self, verbose=False):
        from .metal import FunctionSerializableMetal
        f = FunctionSerializableMetal.build(self.expression, df=self.ds, verbose=verbose, compile=self.ds.is_local())
        function = self.ds.add_function('_jit', f, unique=True)
        return function(*f.arguments)

    def jit_numba(self, verbose=False):
        f = FunctionSerializableNumba.build(self.expression, df=self.ds, verbose=verbose, compile=self.ds.is_local())
        function = self.ds.add_function('_jit', f, unique=True)
        return function(*f.arguments)

    def jit_cuda(self, verbose=False):
        f = FunctionSerializableCuda.build(self.expression, df=self.ds, verbose=verbose, compile=self.ds.is_local())
        function = self.ds.add_function('_jit', f, unique=True)
        return function(*f.arguments)

    def jit_pythran(self, verbose=False):
        import logging
        logger = logging.getLogger('pythran')
        log_level = logger.getEffectiveLevel()
        try:
            if not verbose:
                logger.setLevel(logging.ERROR)
            import pythran
            import imp
            import hashlib
            # self._import_all(module)
            names = []
            funcs = set(expression_namespace.keys())
            expression = self.expression
            if expression in self.ds.virtual_columns:
                expression = self.ds.virtual_columns[self.expression]
            all_vars = self.ds.get_column_names(virtual=True, strings=True, hidden=True) + list(self.ds.variables.keys())
            vaex.expresso.validate_expression(expression, all_vars, funcs, names)
            names = list(set(names))
            types = ", ".join(str(self.ds.data_type(name)) + "[]" for name in names)
            argstring = ", ".join(names)
            code = '''
from numpy import *
#pythran export f({2})
def f({0}):
    return {1}'''.format(argstring, expression, types)
            if verbose:
                print("generated code")
                print(code)
            m = hashlib.md5()
            m.update(code.encode('utf-8'))
            module_name = "pythranized_" + m.hexdigest()
            # print(m.hexdigest())
            module_path = pythran.compile_pythrancode(module_name, code, extra_compile_args=["-DBOOST_SIMD", "-march=native"] + [] if verbose else ["-w"])

            module = imp.load_dynamic(module_name, module_path)
            function_name = "f_" + m.hexdigest()
            function = self.ds.add_function(function_name, module.f, unique=True)

            return Expression(self.ds, "{0}({1})".format(function.name, argstring))
        finally:
                logger.setLevel(log_level)

    def _rename(self, old, new, inplace=False):
        expression = self if inplace else self.copy()
        if old in expression.ast_names:
            for node in expression.ast_names[old]:
                node.id = new
            expression._ast_names[new] = expression._ast_names.pop(old)
        slices = expression._ast_slices
        if old in slices:
            for node in slices[old]:
                if node.value.id == 'df' and isinstance(node.slice.value, ast.Str):
                    node.slice.value.s = new
                else:  # py39
                    node.slice.value = new
        expression._expression = None  # resets the cached string representation
        return expression

    def astype(self, data_type):
        if vaex.array_types.is_string_type(data_type) or data_type == str:
            return self.ds.func.astype(self, 'str')
        else:
            return self.ds.func.astype(self, str(data_type))

    def isin(self, values, use_hashmap=True):
        """Lazily tests if each value in the expression is present in values.

        :param values: List/array of values to check
        :param use_hashmap: use a hashmap or not (especially faster when values contains many elements)
        :return: :class:`Expression` with the lazy expression.
        """
        if self.df.is_category(self) and self.df._future_behaviour:
            labels = self.df.category_labels(self.expression)
            indices = []
            for value in values:
                if value not in labels:
                    pass
                else:
                    indices.append(labels.index(value))
            indices = np.array(indices, dtype=self.index_values().dtype.numpy)
            return self.index_values().isin(indices, use_hashmap=use_hashmap)

        if self.is_string():
            values = pa.array(values, type=pa.large_string())
        else:
             # ensure that values are the same dtype as the expression (otherwise the set downcasts at the C++ level during execution)
            values = np.array(values, dtype=self.dtype.numpy)

        fp_args = vaex.cache.fingerprint(values, use_hashmap)
        fp = f"{self.fingerprint()}_{fp_args}".replace("-", "_")
        if use_hashmap:
            varname = f"var_isin_ordered_set_{fp}"
            if varname not in self.df.variables:
                with self.df._state_lock:
                    if varname not in self.df.variables:
                        # easiest way to create a set is using the vaex dataframe
                        df_values = vaex.from_arrays(x=values)
                        ordered_set = df_values._set(df_values.x)
                        self.df.add_variable(varname, ordered_set)
            return self.df["isin_set(%s, %s)" % (self, varname)]
        else:
            varname = f"var_isin_values_{fp}"
            if varname not in self.df.variables:
                with self.df._state_lock:
                    if varname not in self.df.variables:
                        self.df.add_variable(varname, values)
            return self.df["isin(%s, %s)" % (self, varname)]

    def apply(self, f, vectorize=False, multiprocessing=True):
        """Apply a function along all values of an Expression.

        Shorthand for ``df.apply(f, arguments=[expression])``, see :meth:`DataFrame.apply`

        Example:

        >>> df = vaex.example()
        >>> df.x
        Expression = x
        Length: 330,000 dtype: float64 (column)
        ---------------------------------------
             0  -0.777471
             1    3.77427
             2    1.37576
             3   -7.06738
             4   0.243441

        >>> def func(x):
        ...     return x**2

        >>> df.x.apply(func)
        Expression = lambda_function(x)
        Length: 330,000 dtype: float64 (expression)
        -------------------------------------------
             0   0.604461
             1    14.2451
             2    1.89272
             3    49.9478
             4  0.0592637


        :param f: A function to be applied on the Expression values
        :param vectorize: Call f with arrays instead of a scalars (for better performance).
        :param bool multiprocessing: Use multiple processes to avoid the GIL (Global interpreter lock).
        :returns: A function that is lazily evaluated when called.
        """
        return self.ds.apply(f, [self.expression], vectorize=vectorize, multiprocessing=multiprocessing)

    def dropmissing(self):
        # TODO: df.dropna does not support inplace
        # df = self.df if inplace else self.df.copy()
        df = self.ds
        df = df.dropmissing(column_names=[self.expression])
        return df._expr(self.expression)

    def dropnan(self):
        # TODO: df.dropna does not support inplace
        # df = self.df if inplace else self.df.copy()
        df = self.ds
        df = df.dropnan(column_names=[self.expression])
        return df._expr(self.expression)

    def dropna(self):
        # TODO: df.dropna does not support inplace
        # df = self.df if inplace else self.df.copy()
        df = self.ds
        df = df.dropna(column_names=[self.expression])
        return df._expr(self.expression)

    def map(self, mapper, nan_value=None, missing_value=None, default_value=None, allow_missing=False, axis=None):
        """Map values of an expression or in memory column according to an input
        dictionary or a custom callable function.

        Example:

        >>> import vaex
        >>> df = vaex.from_arrays(color=['red', 'red', 'blue', 'red', 'green'])
        >>> mapper = {'red': 1, 'blue': 2, 'green': 3}
        >>> df['color_mapped'] = df.color.map(mapper)
        >>> df
        #  color      color_mapped
        0  red                   1
        1  red                   1
        2  blue                  2
        3  red                   1
        4  green                 3
        >>> import numpy as np
        >>> df = vaex.from_arrays(type=[0, 1, 2, 2, 2, np.nan])
        >>> df['role'] = df['type'].map({0: 'admin', 1: 'maintainer', 2: 'user', np.nan: 'unknown'})
        >>> df
        #    type  role
        0       0  admin
        1       1  maintainer
        2       2  user
        3       2  user
        4       2  user
        5     nan  unknown
        >>> import vaex
        >>> import numpy as np
        >>> df = vaex.from_arrays(type=[0, 1, 2, 2, 2, 4])
        >>> df['role'] = df['type'].map({0: 'admin', 1: 'maintainer', 2: 'user'}, default_value='unknown')
        >>> df
        #    type  role
        0       0  admin
        1       1  maintainer
        2       2  user
        3       2  user
        4       2  user
        5       4  unknown
        :param mapper: dict like object used to map the values from keys to values
        :param nan_value: value to be used when a nan is present (and not in the mapper)
        :param missing_value: value to use used when there is a missing value
        :param default_value: value to be used when a value is not in the mapper (like dict.get(key, default))
        :param allow_missing: used to signal that values in the mapper should map to a masked array with missing values,
            assumed True when default_value is not None.
        :param bool axis: Axis over which to determine the unique elements (None will flatten arrays or lists)
        :return: A vaex expression
        :rtype: vaex.expression.Expression
        """
        assert isinstance(mapper, collectionsAbc.Mapping), "mapper should be a dict like object"
        if axis is not None:
            raise ValueError('only axis=None is supported')

        df = self.ds
        mapper_keys = list(mapper.keys())
        mapper_values = list(mapper.values())
        try:
            mapper_nan_key_mask = np.isnan(mapper_keys)
        except TypeError:
            # case where we have mixed strings/nan etc
            def try_nan(x):
                try:
                    return np.isnan(x)
                except:
                    return False
            mapper_nan_key_mask = np.array([try_nan(k) for k in mapper_keys])
        mapper_has_nan = mapper_nan_key_mask.sum() > 0
        if mapper_nan_key_mask.sum() > 1:
            raise ValueError('Insanity, you provided multiple nan values as keys for your dict')
        if mapper_has_nan:
            for key, value in mapper.items():
                if key != key:
                    nan_value = value
        for key, value in mapper.items():
            if key is None:
                missing_value = value

        if axis is not None:
            raise ValueError('only axis=None is supported')
        # we map the keys to a ordinal values [0, N-1] using the set
        key_set = df._set(self.expression, flatten=axis is None)
        found_keys = vaex.array_types.tolist(key_set.keys())

        # we want all possible values to be converted
        # so mapper's key should be a superset of the keys found
        use_masked_array = False
        if default_value is not None:
            allow_missing = True
        if allow_missing:
            use_masked_array = True
        if not set(mapper_keys).issuperset(found_keys):
            missing = set(found_keys).difference(mapper_keys)
            missing0 = list(missing)[0]
            only_has_nan = missing0 != missing0 and len(missing) == 1
            if allow_missing:
                if default_value is not None:
                    value0 = list(mapper.values())[0]
                    assert np.issubdtype(type(default_value), np.array(value0).dtype), "default value has to be of similar type"
            else:
                if only_has_nan:
                    pass  # we're good, the hash mapper deals with nan
                else:
                    if missing != {None}:
                        raise ValueError('Missing %i values in mapper: %s' % (len(missing), missing))

        # and these are the corresponding choices
        # note that here we map 'planned' unknown values to the default values
        # and later on in _choose, we map values not even seen in the dataframe
        # to the default_value
        dtype_item = self.data_type(self.expression, axis=-1)
        mapper_keys = dtype_item.create_array(mapper_keys)
        fingerprint = key_set.fingerprint + "-mapper"
        hash_map_unique = vaex.hash.HashMapUnique.from_keys(mapper_keys, fingerprint=fingerprint, dtype=dtype_item)
        indices = hash_map_unique.map(mapper_keys)
        mapper_values = [mapper_values[i] for i in indices]

        choices = [default_value] + [mapper_values[index] for index in indices]
        choices = pa.array(choices)

        key_hash_map_unique_name = df.add_variable('map_key_hash_map_unique', hash_map_unique, unique=True)
        choices_name = df.add_variable('map_choices', choices, unique=True)
        if allow_missing:
            expr = '_map({}, {}, {}, use_missing={!r}, axis={!r})'.format(self, key_hash_map_unique_name, choices_name, use_masked_array, axis)
        else:
            expr = '_map({}, {}, {}, axis={!r})'.format(self, key_hash_map_unique_name, choices_name, axis)
        return Expression(df, expr)

    @property
    def is_masked(self):
        return self.ds.is_masked(self.expression)

    def is_string(self):
        return self.df.is_string(self.expression)


class FunctionSerializable(object):
    pass


@vaex.serialize.register
class FunctionSerializablePickle(FunctionSerializable):
    def __init__(self, f=None, multiprocessing=False):
        self.f = f
        self.multiprocessing = multiprocessing

    def __eq__(self, rhs):
        return self.f == rhs.f

    def pickle(self, function):
        return pickle.dumps(function)

    def unpickle(self, data):
        return pickle.loads(data)

    def __getstate__(self):
        return self.state_get()

    def __setstate__(self, state):
        self.state_set(state)

    def state_get(self):
        data = self.pickle(self.f)
        if vaex.utils.PY2:
            pickled = base64.encodestring(data)
        else:
            pickled = base64.encodebytes(data).decode('ascii')
        return dict(pickled=pickled)

    @classmethod
    def state_from(cls, state, trusted=True):
        obj = cls()
        obj.state_set(state, trusted=trusted)
        return obj

    def state_set(self, state, trusted=True):
        data = state['pickled']
        if vaex.utils.PY2:
            data = base64.decodestring(data)
        else:
            data = base64.decodebytes(data.encode('ascii'))
        if trusted is False:
            raise ValueError("Will not unpickle data when source is not trusted")
        self.f = self.unpickle(data)

    def __call__(self, *args, **kwargs):
        '''Forward the call to the real function'''
        import vaex.multiprocessing
        return vaex.multiprocessing.apply(self._apply, args, kwargs, self.multiprocessing)

    def _apply(self, *args, **kwargs):
        return self.f(*args, **kwargs)


class FunctionSerializableJit(FunctionSerializable):
    def __init__(self, expression, arguments, argument_dtypes, return_dtype, verbose=False, compile=True):
        self.expression = expression
        self.arguments = arguments
        self.argument_dtypes = argument_dtypes
        self.return_dtype = return_dtype
        self.verbose = verbose
        if compile:
            self.f = self.compile()
        else:
            def placeholder(*args, **kwargs):
                raise Exception('You chose not to compile this function (locally), but did invoke it')
            self.f = placeholder

    def state_get(self):
        return dict(expression=self.expression,
                    arguments=self.arguments,
                    argument_dtypes=list(map(lambda dtype: str(dtype.numpy), self.argument_dtypes)),
                    return_dtype=str(self.return_dtype),
                    verbose=self.verbose)

    @classmethod
    def state_from(cls, state, trusted=True):
        return cls(expression=state['expression'],
                   arguments=state['arguments'],
                   argument_dtypes=list(map(lambda s: DataType(np.dtype(s)), state['argument_dtypes'])),
                   return_dtype=DataType(np.dtype(state['return_dtype'])),
                   verbose=state['verbose'])

    @classmethod
    def build(cls, expression, df=None, verbose=False, compile=True):
        df = df or expression.df
        # if it's a virtual column, we probably want to optimize that
        # TODO: fully extract the virtual columns, i.e. depending ones?
        expression = str(expression)

        if expression in df.virtual_columns:
            expression = df.virtual_columns[expression]

        # function validation, and finding variable names
        all_vars = df.get_column_names(hidden=True) + list(df.variables.keys())
        funcs = set(list(expression_namespace.keys()) + list(df.functions.keys()))
        names = []
        vaex.expresso.validate_expression(expression, all_vars, funcs, names)
        # TODO: can we do the above using the Expressio API?s

        arguments = list(set(names))
        argument_dtypes = [df.data_type(argument, array_type='numpy') for argument in arguments]
        return_dtype = df[expression].dtype
        return cls(str(expression), arguments, argument_dtypes, return_dtype, verbose, compile=compile)

    def __call__(self, *args, **kwargs):
        '''Forward the call to the numba function'''
        return self.f(*args, **kwargs)


@vaex.serialize.register
class FunctionSerializableNumba(FunctionSerializableJit):
    def compile(self):
        import numba
        argstring = ", ".join(self.arguments)
        code = '''
from numpy import *
def f({0}):
    return {1}'''.format(argstring, self.expression)
        if self.verbose:
            print('Generated code:\n' + code)
        scope = {}
        exec(code, scope)
        f = scope['f']

        # numba part
        def get_type(name):
            if name == "bool":
                name = "bool_"
            return getattr(numba, name)

        argument_dtypes_numba = [get_type(argument_dtype.numpy.name) for argument_dtype in self.argument_dtypes]
        return_dtype_numba = get_type(self.return_dtype.numpy.name)
        vectorizer = numba.vectorize([return_dtype_numba(*argument_dtypes_numba)])
        return vectorizer(f)


@vaex.serialize.register
class FunctionSerializableCuda(FunctionSerializableJit):
    def compile(self):
        import cupy
        # code generation
        argstring = ", ".join(self.arguments)
        code = '''
from cupy import *
import cupy
@fuse()
def f({0}):
    return {1}
'''.format(argstring, self.expression)#, ";".join(conversions))
        if self.verbose:
            print("generated code")
            print(code)
        scope = dict()#cupy=cupy)
        exec(code, scope)
        func = scope['f']
        def wrapper(*args):
            args = [vaex.array_types.to_numpy(k) for k in args]
            args = [vaex.utils.to_native_array(arg) if isinstance(arg, np.ndarray) else arg for arg in args]
            args = [cupy.asarray(arg) if isinstance(arg, np.ndarray) else arg for arg in args]
            return cupy.asnumpy(func(*args))
        return wrapper


# TODO: this is not the right abstraction, since this won't allow a
# numba version for the function
@vaex.serialize.register
class FunctionToScalar(FunctionSerializablePickle):
    def __call__(self, *args, **kwargs):
        import vaex.multiprocessing
        return vaex.multiprocessing.apply(self._apply, args, kwargs, self.multiprocessing)

    def _apply(self, *args, **kwargs):
        length = len(args[0])
        result = []
        def fix_type(v):
            # TODO: only when column is str type?
            if isinstance(v, np.str_):
                return str(v)
            if isinstance(v, np.bytes_):
                return v.decode('utf8')
            else:
                return v
        args = [vaex.array_types.tolist(k) for k in args]
        for i in range(length):
            scalar_result = self.f(*[fix_type(k[i]) for k in args], **{key: value[i] for key, value in kwargs.items()})
            result.append(scalar_result)
        result = np.array(result)
        return result


class Function(object):

    def __init__(self, dataset, name, f):
        self.dataset = dataset
        self.name = name

        if not vaex.serialize.can_serialize(f): # if not serializable, assume we can use pickle
            f = FunctionSerializablePickle(f)
        self.f = f

    def __call__(self, *args, **kwargs):
        arg_string = ", ".join([str(k) for k in args] + ['{}={:r}'.format(name, value) for name, value in kwargs.items()])
        expression = "{}({})".format(self.name, arg_string)
        return Expression(self.dataset, expression)

class FunctionBuiltin(object):

    def __init__(self, dataset, name, **kwargs):
        self.dataset = dataset
        self.name = name
        self.kwargs = kwargs

    def __call__(self, *args, **kwargs):
        kwargs = dict(kwargs, **self.kwargs)
        arg_string = ", ".join([str(k) for k in args] + ['{}={:r}'.format(name, value) for name, value in kwargs.items()])
        expression = "{}({})".format(self.name, arg_string)
        return Expression(self.dataset, expression)



__author__ = 'maartenbreddels'
import asyncio
import atexit
import traceback
import sys
import aplus
from aplus import listPromise


try:
    # support py 36 for the moment
    import contextvars
    has_contextvars = True
    auto_await_executor = contextvars.ContextVar('auto await executor', default=None)
except ImportError:
    has_contextvars = False



def check_unhandled():
    if Promise.unhandled_exceptions:
        print("Unhandled exceptions in Promises:")
        for exctype, value, tb in Promise.unhandled_exceptions:
            traceback.print_exception(exctype, value, tb)


def rereaise_unhandled():
    if Promise.unhandled_exceptions:
        for exctype, value, tb in Promise.unhandled_exceptions:
            if value:
                raise value.with_traceback(tb)
            # traceback.print_exception(exctype, value, tb)


atexit.register(check_unhandled)


class Promise(aplus.Promise):
    last_exc_info = None
    unhandled_exceptions = []

    @staticmethod
    def create():
        return Promise()

    def create_next(self):
        return Promise()

    @classmethod
    def fulfilled(cls, x):
        p = cls.create()
        p.fulfill(x)
        return p

    @classmethod
    def rejected(cls, reason):
        p = cls.create()
        p.reject(reason)
        return p

    @staticmethod
    def unhandled(exctype, value, traceback):
        # import traceback
        # traceback.print_stack()
        # print "Unhandled error", error
        Promise.unhandled_exceptions.append((exctype, value, traceback))

    def then(self, success=None, failure=None):
        """
        This method takes two optional arguments.  The first argument
        is used if the "self promise" is fulfilled and the other is
        used if the "self promise" is rejected.  In either case, this
        method returns another promise that effectively represents
        the result of either the first of the second argument (in the
        case that the "self promise" is fulfilled or rejected,
        respectively).

        Each argument can be either:
          * None - Meaning no action is taken
          * A function - which will be called with either the value
                of the "self promise" or the reason for rejection of
                the "self promise".  The function may return:
                * A value - which will be used to fulfill the promise
                  returned by this method.
                * A promise - which, when fulfilled or rejected, will
                  cascade its value or reason to the promise returned
                  by this method.
          * A value - which will be assigned as either the value
                or the reason for the promise returned by this method
                when the "self promise" is either fulfilled or rejected,
                respectively.

        :type success: (object) -> object
        :type failure: (object) -> object
        :rtype : Promise
        """
        ret = self.create_next()

        def callAndFulfill(v):
            """
            A callback to be invoked if the "self promise"
            is fulfilled.
            """
            try:
                if aplus._isFunction(success):
                    ret.fulfill(success(v))
                else:
                    ret.fulfill(v)
            except Exception as e:
                Promise.last_exc_info = sys.exc_info()
                e.exc_info = sys.exc_info()
                ret.reject(e)

        def callAndReject(r):
            """
            A callback to be invoked if the "self promise"
            is rejected.
            """
            try:
                if aplus._isFunction(failure):
                    ret.fulfill(failure(r))
                else:
                    ret.reject(r)
            except Exception as e:
                Promise.last_exc_info = sys.exc_info()
                e.exc_info = sys.exc_info()
                ret.reject(e)

        self.done(callAndFulfill, callAndReject)

        return ret

    def end(self):
        def failure(reason):
            args = sys.exc_info()
            if args is None or args[0] is None:
                args = Promise.last_exc_info
            # print reason, args, Promise.last_exc_info, sys.exc_info(), Promise.unhandled
            if hasattr(reason, "exc_info"):
                args = reason.exc_info
            # print args
            try:
                Promise.unhandled(*args)
            except:
                print("Error in unhandled handler")
                traceback.print_exc()
        return self.then(None, failure)


    def __await__(self):
        return self._create_awaitable().__await__()

    async def _create_awaitable(self):
        future = asyncio.Future()
        self.then(future.set_result, future.set_exception)
        if has_contextvars:
            executor = auto_await_executor.get()
            if executor:
                result = await asyncio.gather(executor.execute_async(), future)
                return result[-1]
        return await future


aplus.Promise = Promise



def alias_main(argv):
    import argparse
    parser = argparse.ArgumentParser(argv[0])
    # parser.add_argument('--verbose', '-v', action='count', default=0)
    # parser.add_argument('--quiet', '-q', default=False, action='store_true', help="do not output anything")
    # parser.add_argument('--list', '-l', default=False, action='store_true', help="list columns of input")
    # parser.add_argument('--progress', help="show progress (default: %(default)s)", default=True, action='store_true')
    # parser.add_argument('--no-progress', dest="progress", action='store_false')
    # parser.add_argument('--shuffle', "-s", dest="shuffle", action='store_true', default=False)

    subparsers = parser.add_subparsers(help='type of task', dest="task")

    parser_list = subparsers.add_parser('list', help='list aliases')

    parser_add = subparsers.add_parser('add', help='add alias')
    parser_add.add_argument('name', help='name of alias')
    parser_add.add_argument('path', help='path/filename for alias')
    parser.add_argument('-f', '--force', help="force/overwrite existing alias", default=False, action='store_true')

    parser_remove = subparsers.add_parser('remove', help='remove alias')
    parser_remove.add_argument('name', help='name of alias')

    args = parser.parse_args(argv[1:])
    import vaex
    if args.task == "add":
        vaex.aliases[args.name] = args.path
    if args.task == "remove":
        del vaex.aliases[args.name]
    if args.task == "list":
        for name in sorted(vaex.aliases.keys()):
            print("%s: %s" % (name, vaex.aliases[name]))


def make_stat_parser(name):
    import argparse
    parser = argparse.ArgumentParser(name)
    # parser.add_argument('--verbose', '-v', action='count', default=0)
    # parser.add_argument('--quiet', '-q', default=False, action='store_true', help="do not output anything")
    # parser.add_argument('--list', '-l', default=False, action='store_true', help="list columns of input")
    # parser.add_argument('--progress', help="show progress (default: %(default)s)", default=True, action='store_true')
    # parser.add_argument('--no-progress', dest="progress", action='store_false')
    # parser.add_argument('--shuffle', "-s", dest="shuffle", action='store_true', default=False)

    # subparsers = parser.add_subparsers(help='type of task', dest="task")

    # parser_list = subparsers.add_parser('list', help='list aliases')

    # parser = subparsers.add_parser('add', help='add alias')
    parser.add_argument('dataset', help='path or name of dataset')
    parser.add_argument('--fraction', "-f", dest="fraction", type=float, default=1.0, help="fraction of input dataset to export")
    return parser


def stat_main(argv):
    parser = make_stat_parser(argv[0])
    args = parser.parse_args(argv[1:])
    import vaex
    dataset = vaex.open(args.dataset)
    if dataset is None:
        print("Cannot open input: %s" % args.dataset)
        sys.exit(1)
    print("dataset:")
    print("  length: %s" % len(dataset))
    print("  full_length: %s" % dataset.full_length())
    print("  name: %s" % dataset.name)
    print("  path: %s" % dataset.path)
    print("  columns: ")
    desc = dataset.description
    if desc:
        print("    description: %s" % desc)
    for name in dataset.get_column_names():
        print("   - %s: " % name)
        desc = dataset.descriptions.get(name)
        if desc:
            print("  \tdescription: %s" % desc)
        unit = dataset.unit(name)
        if unit:
            print("   \tunit: %s" % unit)
        dtype = dataset.data_type(name)
        print("   \ttype: %s" % dtype.name)



from typing import Sequence
import dask.base
import numpy as np
import pyarrow as pa

import vaex
import vaex.array_types

class DataType:
    """Wraps numpy and arrow data types in a uniform interface

    Examples:
    >>> import numpy as np
    >>> import pyarrow as pa
    >>> type1 = DataType(np.dtype('f8'))
    >>> type1
    float64
    >>> type2 = DataType(np.dtype('>f8'))
    >>> type2
    >f8
    >>> type1 in [float, int]
    True
    >>> type1 == type2
    False
    >>> type1 == pa.float64()
    True
    >>> type1 == pa.int64()
    False
    >>> DataType(np.dtype('f4'))
    float32
    >>> DataType(pa.float32())
    float32

    """
    def __init__(self, dtype):
        if isinstance(dtype, DataType):
            self.internal = dtype.internal
        else:
            if isinstance(dtype, pa.DataType):
                self.internal = dtype
            else:
                self.internal = np.dtype(dtype)

    def to_native(self):
        '''Removes non-native endianness'''
        return DataType(vaex.utils.to_native_dtype(self.internal))        

    def __hash__(self):
        return hash((self.__class__.__name__, self.internal))

    def __eq__(self, other):
        if self.is_encoded:
            return self.value_type == other
        if other is str:
            return self.is_string
        if other is float:
            return self.is_float
        if other is int:
            return self.is_integer
        if other is list:
            return self.is_list
        if other is dict:
            return self.is_struct
        if other is object:
            return self.is_object
        if isinstance(other, str):
            tester = 'is_' + other
            if hasattr(self, tester):
                return getattr(self, tester)
        if not isinstance(other, DataType):
            other = DataType(other)
        if other.is_primitive:
            if self.is_arrow:
                other = DataType(other.arrow)
            if self.is_numpy:
                other = DataType(other.numpy)
        return vaex.array_types.same_type(self.internal, other.internal)

    def __repr__(self):
        '''Standard representation for datatypes


        >>> dtype = DataType(pa.float64())
        >>> dtype.internal
        DataType(double)
        >>> dtype
        float64
        >>> DataType(pa.float32())
        float32
        >>> DataType(pa.dictionary(pa.int32(), pa.string()))
        dictionary<values=string, indices=int32, ordered=0>
        '''

        internal = self.internal
        if self.is_datetime:
            internal = self.numpy

        repr = str(internal)
        translate = {
            'datetime64': 'datetime64[ns]',  # for consistency, always add time unit
            'double': 'float64',  # this is what arrow does
            'float': 'float32',  # this is what arrow does
        }
        return translate.get(repr, repr)

    @property
    def name(self):
        '''Alias of dtype.numpy.name or str(dtype.arrow) if not primitive

        >>> DataType(np.dtype('f8')).name
        'float64'
        >>> DataType(np.dtype('>f4')).name
        'float32'
        >>> DataType(pa.float64()).name
        'float64'
        >>> DataType(pa.large_string()).name
        'large_string'
        >>> DataType(pa.string()).name
        'string'
        >>> DataType(pa.bool_()).name
        'bool'
        >>> DataType(np.dtype('?')).name
        'bool'
        >>> DataType(pa.dictionary(pa.int32(), pa.string())).name
        'dictionary<values=string, indices=int32, ordered=0>'
        '''
        return self.numpy.name if (self.is_primitive or self.is_datetime) else str(self.internal)

    @property
    def kind(self):
        return self.numpy.kind

    @property
    def numpy(self):
        '''Return the numpy equivalent type

        >>> DataType(pa.float64()).numpy == np.dtype('f8')
        True
        '''
        return vaex.array_types.to_numpy_type(self.internal)

    @property
    def arrow(self):
        '''Return the Apache Arrow equivalent type

        >>> DataType(np.dtype('f8')).arrow == pa.float64()
        True
        '''
        return vaex.array_types.to_arrow_type(self.internal)

    @property
    def is_arrow(self):
        '''Return True if it wraps an Arrow type

        >>> DataType(pa.string()).is_arrow
        True
        >>> DataType(pa.int32()).is_arrow
        True
        >>> DataType(np.dtype('f8')).is_arrow
        False
        '''
        return isinstance(self.internal, pa.DataType)

    @property
    def is_numpy(self):
        '''Return True if it wraps an NumPy dtype

        >>> DataType(np.dtype('f8')).is_numpy
        True
        >>> DataType(pa.string()).is_numpy
        False
        >>> DataType(pa.int32()).is_numpy
        False
        '''
        return isinstance(self.internal, np.dtype)

    @property
    def is_numeric(self):
        '''Tests if type is numerical (float, int)

        >>> DataType(np.dtype('f8')).is_numeric
        True
        >>> DataType(pa.float32()).is_numeric
        True
        >>> DataType(pa.large_string()).is_numeric
        False
        '''
        try:
            return self.kind in 'fiu'
        except NotImplementedError:
            return False

    @property
    def is_primitive(self):
        '''Tests if type is numerical (float, int, bool)

        >>> DataType(np.dtype('b')).is_primitive
        True
        >>> DataType(pa.bool_()).is_primitive
        True
        '''
        if self.is_arrow:
            return pa.types.is_primitive(self.internal)
        else:
            return self.kind in 'fiub'

    @property
    def is_datetime(self):
        """Tests if dtype is datetime (numpy) or timestamp (arrow)

        Date/Time:
        >>> date_type = DataType(np.dtype('datetime64'))
        >>> date_type
        datetime64[ns]
        >>> date_type == 'datetime'
        True

        Using Arrow:

        >>> date_type = DataType(pa.timestamp('ns'))
        >>> date_type
        datetime64[ns]
        >>> date_type == 'datetime'
        True
        >>> date_type = DataType(pa.large_string())
        >>> date_type.is_datetime
        False
        """
        if self.is_arrow:
            return pa.types.is_timestamp(self.internal)
        else:
            return self.kind in 'M'

    @property
    def is_timedelta(self):
        '''Test if timedelta

        >>> dtype = DataType(np.dtype('timedelta64'))
        >>> dtype
        timedelta64
        >>> dtype == 'timedelta'
        True
        >>> dtype.is_timedelta
        True
        >>> date_type = DataType(pa.large_string())
        >>> date_type.is_timedelta
        False
        '''
        if self.is_arrow:
            return isinstance(self.arrow, pa.DurationType)
        else:
            return self.kind in 'm'

    @property
    def is_temporal(self):
        '''Alias of (is_datetime or is_timedelta)'''
        return self.is_datetime or self.is_timedelta

    @property
    def is_float(self):
        '''Test if a float (float32 or float64)

        >>> dtype = DataType(np.dtype('float32'))
        >>> dtype
        float32
        >>> dtype == 'float'
        True
        >>> dtype == float
        True
        >>> dtype.is_float
        True

        Using Arrow:
        >>> DataType(pa.float32()) == float
        True
        '''
        return self.is_primitive and vaex.array_types.to_numpy_type(self.internal).kind in 'f'

    @property
    def is_unsigned(self):
        '''Test if an (unsigned) integer

        >>> dtype = DataType(np.dtype('uint32'))
        >>> dtype
        uint32
        >>> dtype == 'unsigned'
        True
        >>> dtype.is_unsigned
        True

        Using Arrow:
        >>> DataType(pa.uint32()).is_unsigned
        True
        '''
        return self.is_primitive and vaex.array_types.to_numpy_type(self.internal).kind in 'u'

    @property
    def is_signed(self):
        '''Test if a (signed) integer

        >>> dtype = DataType(np.dtype('int32'))
        >>> dtype
        int32
        >>> dtype == 'signed'
        True

        Using Arrow:
        >>> DataType(pa.int32()).is_signed
        True
        '''
        return self.is_primitive and vaex.array_types.to_numpy_type(self.internal).kind in 'i'

    @property
    def is_integer(self):
        '''Test if an (unsigned or signed) integer

        >>> DataType(np.dtype('uint32')) == 'integer'
        True
        >>> DataType(np.dtype('int8')) == int
        True
        >>> DataType(np.dtype('int16')).is_integer
        True

        Using Arrow:
        >>> DataType(pa.uint32()).is_integer
        True
        >>> DataType(pa.int16()) == int
        True

        '''
        return self.is_primitive and vaex.array_types.to_numpy_type(self.internal).kind in 'iu'

    @property
    def is_string(self):
        '''Test if an (arrow) string or large_string

        >>> DataType(pa.string()) == str
        True
        >>> DataType(pa.large_string()) == str
        True
        >>> DataType(pa.large_string()).is_string
        True
        >>> DataType(pa.large_string()) == 'string'
        True
        '''
        return vaex.array_types.is_string_type(self.internal)

    @property
    def is_list(self):
        '''Test if an (arrow) list or large_string

        >>> DataType(pa.list_(pa.string())) == list
        True
        >>> DataType(pa.large_list(pa.string())) == list
        True
        >>> DataType(pa.list_(pa.string())).is_list
        True
        >>> DataType(pa.list_(pa.string())) == 'list'
        True
        '''
        return self.is_arrow and (pa.types.is_list(self.internal) or pa.types.is_large_list(self.internal))

    @property
    def is_struct(self) -> bool:
        '''Test if an (arrow) struct

        >>> DataType(pa.struct([pa.field('a', pa.utf8())])) == dict
        True
        >>> DataType(pa.struct([pa.field('a', pa.utf8())])).is_struct
        True
        >>> DataType(pa.struct([pa.field('a', pa.utf8())])) == 'struct'
        True
        '''
        return self.is_arrow and pa.types.is_struct(self.internal)

    @property
    def is_object(self):
        '''Test if a NumPy dtype=object (avoid if possible)'''
        return self.is_numpy and self.internal == object

    @property
    def is_encoded(self):
        '''Test if an (arrow) dictionary type (encoded data)

        >>> DataType(pa.dictionary(pa.int32(), pa.string())) == str
        True
        >>> DataType(pa.dictionary(pa.int32(), pa.string())).is_encoded
        True
        '''
        return self.is_arrow and (pa.types.is_dictionary(self.internal))

    @property
    def value_type(self):
        '''Return the DataType of the list values or values of an encoded type

        >>> DataType(pa.list_(pa.string())).value_type
        string
        >>> DataType(pa.list_(pa.float64())).value_type
        float64
        >>> DataType(pa.dictionary(pa.int32(), pa.string())).value_type
        string
        '''
        if not (self.is_list or self.is_encoded):
            raise TypeError( f'{self} is not a list or encoded type')
        return DataType(self.internal.value_type)

    @property
    def index_type(self):
        '''Return the DataType of the index of an encoded type, or simple the type

        >>> DataType(pa.string()).index_type
        string
        >>> DataType(pa.dictionary(pa.int32(), pa.string())).index_type
        int32
        '''
        type = self.internal
        if self.is_encoded:
            type = self.internal.index_type
        return DataType(type)

    def upcast(self):
        '''Cast to the higest data type matching the type

        >>> DataType(np.dtype('uint32')).upcast()
        uint64
        >>> DataType(np.dtype('int8')).upcast()
        int64
        >>> DataType(np.dtype('float32')).upcast()
        float64

        Using Arrow
        >>> DataType(pa.float32()).upcast()
        float64
        '''
        return DataType(vaex.array_types.upcast(self.internal))

    @property
    def byteorder(self):
        return self.numpy.byteorder

    def create_array(self, values : Sequence):
        '''Create an array from a sequence with the same dtype

        If values is a list containing None, it will map to a masked array (numpy) or null values (arrow)

        >>> DataType(np.dtype('float32')).create_array([1., 2.5, None, np.nan])
        masked_array(data=[1.0, 2.5, --, nan],
                     mask=[False, False,  True, False],
               fill_value=1e+20)
        >>> DataType(pa.float32()).create_array([1., 2.5, None, np.nan])  # doctest:+ELLIPSIS
        <pyarrow.lib.FloatArray object at ...>
        [
          1,
          2.5,
          null,
          nan
        ]
        '''
        if self.is_arrow:
            if vaex.array_types.is_arrow_array(values):
                return values
            else:
                return pa.array(values, type=self.arrow)
        else:
            if isinstance(values, np.ndarray):
                return values.astype(self.internal, copy=False)
            mask = [k is None for k in values]
            if any(mask):
                values = [values[0] if k is None else k for k in values]
                return np.ma.array(values, mask=mask)
            else:
                return np.array(values)
            return np.asarray(values, dtype=self.numpy)

@dask.base.normalize_token.register(DataType)
def normalize_DataType(t):
    return type(t).__name__, t.internal



"""This module contains the `register_function` decorator to add expression methods to vaex dataframe."""

import functools

import vaex.arrow
import vaex.expression
import vaex.multiprocessing

scopes = {
    'str': vaex.expression.StringOperations,
    'str_pandas': vaex.expression.StringOperationsPandas,
    'dt': vaex.expression.DateTime,
    'td': vaex.expression.TimeDelta,
    'struct': vaex.expression.StructOperations
}


def register_function(scope=None, as_property=False, name=None, on_expression=True, df_accessor=None,
                      multiprocessing=False):
    """Decorator to register a new function with vaex.

    If on_expression is True, the function will be available as a method on an
    Expression, where the first argument will be the expression itself.

    If `df_accessor` is given, it is added as a method to that dataframe accessor (see e.g. vaex/geo.py)

    Example:

    >>> import vaex
    >>> df = vaex.example()
    >>> @vaex.register_function()
    >>> def invert(x):
    >>>     return 1/x
    >>> df.x.invert()


    >>> import numpy as np
    >>> df = vaex.from_arrays(departure=np.arange('2015-01-01', '2015-12-05', dtype='datetime64'))
    >>> @vaex.register_function(as_property=True, scope='dt')
    >>> def dt_relative_day(x):
    >>>     return vaex.functions.dt_dayofyear(x)/365.
    >>> df.departure.dt.relative_day
    """
    import vaex.multiprocessing
    prefix = ''
    if scope:
        prefix = scope + "_"
        if scope not in scopes:
            raise KeyError("unknown scope")

    def wrapper(f, name=name):
        name = name or f.__name__
        # remove possible prefix
        if name.startswith(prefix):
            name = name[len(prefix):]
        full_name = prefix + name
        if df_accessor:
            def closure(name=name, full_name=full_name, function=f):
                def wrapper(self, *args, **kwargs):
                    lazy_func = getattr(self.df.func, full_name)
                    lazy_func = vaex.arrow.numpy_dispatch.autowrapper(lazy_func)
                    return vaex.multiprocessing.apply(lazy_func, args, kwargs, multiprocessing)

                return functools.wraps(function)(wrapper)

            if as_property:
                setattr(df_accessor, name, property(closure()))
            else:
                setattr(df_accessor, name, closure())
        else:
            if on_expression:
                if scope:
                    def closure(name=name, full_name=full_name, function=f):
                        def wrapper(self, *args, **kwargs):
                            lazy_func = getattr(self.expression.ds.func, full_name)
                            lazy_func = vaex.arrow.numpy_dispatch.autowrapper(lazy_func)
                            args = (self.expression,) + args
                            return vaex.multiprocessing.apply(lazy_func, args, kwargs, multiprocessing)

                        return functools.wraps(function)(wrapper)

                    if as_property:
                        setattr(scopes[scope], name, property(closure()))
                    else:
                        setattr(scopes[scope], name, closure())
                else:
                    def closure(name=name, full_name=full_name, function=f):
                        def wrapper(self, *args, **kwargs):
                            lazy_func = getattr(self.ds.func, full_name)
                            lazy_func = vaex.arrow.numpy_dispatch.autowrapper(lazy_func)
                            args = (self,) + args
                            return vaex.multiprocessing.apply(lazy_func, args, kwargs, multiprocessing=multiprocessing)

                        return functools.wraps(function)(wrapper)

                    setattr(vaex.expression.Expression, name, closure())
        vaex.expression.expression_namespace[prefix + name] = vaex.arrow.numpy_dispatch.autowrapper(f)
        return f  # we leave the original function as is

    return wrapper



from typing import List
import warnings

from .utils import get_env_type, RegistryCallable
import vaex
import vaex.settings

_progressbar_registry = RegistryCallable("vaex.progressbar", "progressbar")


# these are registered in the entry_points
def simple(type=None, title="processing", max_value=1):
    import vaex.misc.progressbar as pb
    return pb.ProgressBar(0, 1, title=title)

def widget(type=None, title="processing", max_value=1):
    import vaex.misc.progressbar as pb
    return pb.ProgressBarWidget(0, 1, title=title)

def rich(type=None, title="processing", max_value=1):
    import vaex.misc.progressbar as pb
    return pb.ProgressBarRich(0, 1, title=title)


_last_progress_tree = None
_last_traceback = None
_progress_tree_stack = []

class ProgressTree:
    def __init__(self, children=None, next=None, bar=None, parent=None, name=None, hide=False):
        global _last_progress_tree, _last_traceback
        self.next = next
        self.children : List[ProgressTree] = children or list()
        self.finished = False
        self.last_fraction = None
        self.fraction = 0
        self.root = self if parent is None else parent.root
        self.passes_start = None
        self.bar = bar
        self.parent = parent
        self.name = name
        self.hide = hide
        self.cancelled = False
        self.oncancel = lambda: None
        if self.bar:
            self.bar.start()
        if parent is None and not hide:
            if _last_progress_tree is None:
                _last_progress_tree = self
                import traceback
                _last_traceback = ''.join(traceback.format_stack(limit=8))
            else:
                if vaex.DEBUG_MODE:
                    warnings.warn("Already a root progress tree exists, you may want to pass down the progress argument", stacklevel=4)
                    import traceback
                    trace = ''.join(traceback.format_stack(limit=7))
                    print('progress tree triggerd from:\n', trace)
                    print("====\nPrevious traceback:\n", _last_traceback)

    def cancel(self):
        self.cancelled = True

    def __repr__(self):
        name = self.__class__.__module__ + "." + self.__class__.__name__
        return "<%s(name=%r)> instance at 0x%x" % (name, self.name, id(self))

    def __enter__(self):
        _progress_tree_stack.append(self)
        if self.bar:
            self.bar.start()
        return self

    def __exit__(self, *args):
        _progress_tree_stack.pop()
        self.exit()

    def exit(self):
        global _last_progress_tree
        if self.parent is None:
            _last_progress_tree = None
        self(1)
        if self.bar:
            self.bar.exit()

    def exit_on(self, promise):
        def ok(arg):
            self.exit()
            return arg
        def error(arg):
            self.exit()
            raise arg
        return promise.then(ok, error)
        # return promise


    def hidden(self):
        '''Will add a child node that is hidden and not participate in progress calculations'''
        pb = ProgressTree(hide=True)
        return pb

    def add(self, name=None):
        pb = ProgressTree(parent=self, name=name)
        if self.bar and hasattr(self.bar, 'add_child') and not self.hide:
            pb.bar = self.bar.add_child(pb, None, name)
        self.children.append(pb)
        self.finished = False
        self.fraction = sum([c.fraction for c in self.children]) / len(self.children)
        self(self.fraction)
        if self.root.bar:
            # could be that it was stopped
            self.root.bar.start()

        return pb

    def add_task(self, task, name=None, status_ready="from cache"):
        pb = self.add(name)
        pb.oncancel = task.cancel
        task.signal_progress.connect(pb)
        if task.isFulfilled:
            pb(status_ready)
            pb(1)
        def on_error(exc):
            if pb.bar:
                pb.bar.set_status(str(exc))
                pb(pb.last_fraction)
                pb.finished = True
                if pb.root is not pb:
                    pb.root(None)
        def on_start(executor):
            passes = executor.passes
            if self.root.passes_start is None:
                self.root.passes_start = passes
            if pb.bar:
                pb.bar.set_passes(passes - self.root.passes_start + 1)
        task.signal_start.connect(on_start)
        task.then(None, on_error).end()
        return pb

    def __call__(self, fraction_or_status):
        if isinstance(fraction_or_status, str):
            self.status = fraction_or_status
            if self.bar:
                self.bar.set_status(self.status)
            return True
        fraction = fraction_or_status
        if fraction != 1.0:
            self.finished = False
        if self.cancelled:
            return False
        # ignore fraction
        result = True
        if len(self.children) == 0:
            self.fraction = fraction
        else:
            self.fraction = sum([c.fraction if not c.finished else 1 for c in self.children]) / len(self.children)
        fraction = self.fraction
        if fraction != self.last_fraction:  # avoid too many calls
            if fraction == 1 and not self.finished:  # make sure we call finish only once
                self.finished = True
                if self.bar:
                    self.bar.finish()
            elif fraction != 1:
                if self.bar:
                    self.bar.update(fraction)
        if self.next:
            result = self.next(fraction)
        if self.parent:
            assert self in self.parent.children
            result = self.parent(None) in [None, True] and result  # fraction is not used anyway..
            if result is False:
                self.oncancel()
        self.last_fraction = fraction
        return result


def bar(type_name=None, title="processing", max_value=1):
    if type_name is None:
        type_name = vaex.settings.main.progress.force
        if type_name is None:
            type_name = vaex.settings.main.progress.type
    return _progressbar_registry[type_name](title=title)


def tree(f=None, next=None, name=None, title=None):
    if f in [False, None] and _progress_tree_stack:
        f = _progress_tree_stack[-1]
    if f in [False, None] and vaex.settings.main.progress.force:
        f = vaex.settings.main.progress.force
    if name is None:
        name = title
    if title is None:
        name = title
    if isinstance(f, ProgressTree):
        # collapse nameless or similarly named progressbars in 1
        if title is None or f.bar and f.bar.title == title:
            return f
        else:
            return f.add(title)
    if callable(f):
        next = f
        f = False
    if f in [None, False]:
        return ProgressTree(next=next, name=name)
    else:
        if f is True:
            return ProgressTree(bar=bar(title=title), next=next, name=name)
        elif isinstance(f, str):
            return ProgressTree(bar=bar(f, title=title), next=next, name=name)
        else:
            return ProgressTree(next=next, name=name)



from abc import ABC, abstractmethod
import logging
import os
import warnings

import six
import numpy as np
import pyarrow as pa
import pyarrow.compute as pc

import vaex
import vaex.utils
import vaex.cache
from .array_types import supported_array_types, supported_arrow_array_types, string_types, is_string_type


if vaex.utils.has_c_extension:
    import vaex.strings

logger = logging.getLogger("vaex.column")


register = vaex.encoding.make_class_registery('column')

class Column(ABC):
    _cached_fingerprint = None

    def tolist(self):
        return self.to_numpy().tolist()

    def fingerprint(self):
        '''id that uniquely identifies a column cross runtime'''
        if self._cached_fingerprint is None:
            self._cached_fingerprint = self._fingerprint()
        return self._cached_fingerprint

    @abstractmethod
    def _fingerprint(self):
        raise NotImplementedError

    def to_arrow(self, type=None):
        return pa.array(self, type=type)

supported_column_types = (Column, ) + supported_array_types

def is_column_like(col):
    return isinstance(col, supported_column_types)


class ColumnVirtualRange(Column):
    def __init__(self, start, stop, step=1, dtype=None):
        self.start = start
        self.stop = stop
        self.step = step
        self.dtype = np.dtype(dtype)
        self.shape = (self.__len__(),)

    def __len__(self):
        return (self.stop - self.start) // self.step

    def _fingerprint(self):
        fp = vaex.cache.fingerprint(self.start, self.stop, self.step, self.dtype, self.shape)
        return f'column-vrange-{fp}'

    def __getitem__(self,  slice):
        start, stop, step = slice.start, slice.stop, slice.step
        if start is None:
            start = self.start
        if stop is None:
            stop = self.stop
        if step is None:
            step = self.step
        return np.arange(self.start + start, self.start + stop, step, dtype=self.dtype)

    def trim(self, i1, i2):
        return ColumnVirtualRange(self.start + i1 * self.step, self.start + i2 * self.step, self.step, self.dtype)


class ColumnVirtualConstant(Column):
    def __init__(self, value, length, dtype=None, chunk_size=1024):
        self.length = length
        self.dtype = self._get_dtype(value=value, dtype=dtype)
        self.chunk_size = chunk_size
        self.value = self._get_value(value=value)

    def __len__(self):
        return self.length

    def __getitem__(self, item):
        template = pa.array([self.value] * self.chunk_size, type=self.dtype.arrow)
        n_chunks = (self.length + self.chunk_size - 1) // self.chunk_size
        if item.step not in [None, 1]:
            raise ValueError('step can only be 1')
        start, stop, step = item.start, item.stop, item.step
        start = start or 0
        stop = stop or len(self)
        assert step in [None, 1]
        item = slice(start, stop, step)
        ar = pa.chunked_array([template] * n_chunks)[item]
        return ar

    def _fingerprint(self):
        fp = vaex.cache.fingerprint(self.value, self.length, self.dtype, self.chunk_size)
        return f'column-vconstant-{fp}'

    def _get_dtype(self, value, dtype):
        if dtype is None:
            return vaex.dtype(pa.array([value]).type)
        else:
            if isinstance(value, list):
                return vaex.dtype(pa.list_(vaex.dtype(dtype).arrow))
            return vaex.dtype(dtype)

    def _get_value(self, value):
        return pc.cast(pa.scalar(value), self.dtype.arrow).as_py()

    def trim(self, i1, i2):
        length = i2 - i1
        return ColumnVirtualConstant(value=self.value, length=length)


class ColumnMaskedNumpy(Column):
    def __init__(self, data, mask):
        self.data = data
        self.mask = mask
        self.dtype = data.dtype
        assert len(data) == len(mask)

    def _fingerprint(self):
        hash1 = vaex.dataset.hash_array_data(self.data[:])
        hash2 = vaex.dataset.hash_array_data(self.indimaskces[:])
        fp = vaex.cache.fingerprint(hash1, hash2)
        return f'column-masked-numpy-{fp}'

    def __len__(self):
        return len(self.data)

    def __getitem__(self,  slice):
        data = self.data[slice]
        mask = self.mask[slice]
        return np.ma.array(data, mask=mask, shrink=False)

    def trim(self, i1, i2):
        return ColumnMaskedNumpy(self.data.trim(i1, i2), self.mask.trim(i1, i2))


class ColumnSparse(Column):
    def __init__(self, matrix, column_index):
        self.matrix = matrix
        self.column_index = column_index
        self.shape = self.matrix.shape[:1]
        self.dtype = self.matrix.dtype

    def _fingerprint(self):
        fp = vaex.cache.fingerprint(self.matrix, self.column_index)
        return f'column-sparse-{fp}'

    def __len__(self):
        return self.shape[0]

    def trim(self, i1, i2):
        return ColumnSparse(self.matrix[i1:i2], column_index=self.column_index)

    def __getitem__(self, slice):
        # not sure if this is the fastest
        return self.matrix[slice, self.column_index].A[:,0]


class ColumnNumpyLike(Column):
    """Wraps a numpy like object (like hdf5 dataset) into a column vaex understands"""
    def __init__(self, ar):
        self.ar = ar  # this should behave like a numpy array
        self.dtype = ar.dtype

    def __len__(self):
        return len(self.ar)

    def trim(self, i1, i2):
        return type(self)(self.ar[i1:i2])

    def __getitem__(self, slice):
        return self.ar[slice]

    def __setitem__(self, slice, value):
        self.ar[slice] = value

    def _fingerprint(self):
        ar = self[:]
        hash = vaex.dataset.hash_array(ar)
        fp = vaex.cache.fingerprint(hash)
        return f'column-numpy-like-{fp}'


@register
class ColumnArrowLazyCast(Column):
    snake_name = "lazy_cast"
    """Wraps an array like object and cast it lazily"""
    def __init__(self, ar, type):
        self.ar = ar  # this should behave like a numpy array
        self.type = vaex.dtype(type)

    def encode(self, encoding):
        return {
            'array': encoding.encode('array', self.ar),
            'dtype': encoding.encode('dtype', self.type),
        }

    @classmethod
    def decode(cls, encoding, spec):
        return cls(encoding.decode('array', spec['array']),
                   encoding.decode('dtype', spec['dtype']))

    def __arrow_array__(self, type=None):
        return pa.array(self.ar)

    def _fingerprint(self):
        hash = vaex.dataset.hash_array_data(self.ar)
        fp = vaex.cache.fingerprint(hash, self.type)
        return f'column-arrow-lazy-cast-{fp}'

    @property
    def nbytes(self):
        return self.ar.nbytes

    @property
    def dtype(self):
        return self.type

    def __len__(self):
        return len(self.ar)

    def trim(self, i1, i2):
        return type(self)(self.ar[i1:i2], self.type)

    def __getitem__(self, slice):
        if self.ar.dtype == object and vaex.array_types.is_string_type(self.type):
            # this seem to be the only way to convert mixed str and nan to include nulls
            return pa.Array.from_pandas(self.ar[slice], type=self.type.arrow)
        return pa.array(self.ar[slice], type=self.type.arrow)


@register
class ColumnArrowDictionaryEncoded(Column):
    snake_name = "dictionary_encoded"
    """Wraps an array like object and cast it lazily to arrow dict encoded"""
    def __init__(self, indices, dictionary):
        self.indices = indices  # should be column like
        self.dictionary = vaex.array_types.to_arrow(dictionary)  # lets do this once for performance

    def encode(self, encoding):
        return {
            'indices': encoding.encode('array', self.indices),
            'dictionary': encoding.encode('array', self.dictionary),
        }

    @classmethod
    def decode(cls, encoding, spec):
        return cls(encoding.decode('array', spec['indices']),
                   encoding.decode('array', spec['dictionary']))

    def __arrow_array__(self, type=None):
        return self[:]

    def _fingerprint(self):
        hash1 = vaex.dataset.hash_array_data(self.indices)
        hash2 = vaex.dataset.hash_array_data(self.dictionary)
        fp = vaex.cache.fingerprint(hash1, hash2)
        return f'column-arrow-{self.snake_name}-{fp}'

    @property
    def nbytes(self):
        # consistent with arrow
        if int(pa.__version__.split(".")[0]) >= 7:
            return self.indices.nbytes + self.dictionary.nbytes
        else:
            return self.indices.nbytes

    @property
    def dtype(self):
        return self[0:0].type

    def __len__(self):
        return len(self.indices)

    def trim(self, i1, i2):
        return type(self)(self.indices[i1:i2], self.dictionary)

    def __getitem__(self, slice):
        return pa.DictionaryArray.from_arrays(self.indices[slice], self.dictionary, safe=False)


class ColumnIndexed(Column):
    def __init__(self, column, indices, masked=False):
        self.column = column
        self.indices = indices
        self.dtype = vaex.array_types.data_type(column)
        self.shape = (len(indices),)
        self.masked = masked
        # this check is too expensive
        # max_index = self.indices.max()
        # if not np.ma.is_masked(max_index):
        #     assert max_index < self.df._length_original

    def _fingerprint(self):
        hash1 = vaex.dataset.hash_array_data(self.column[:])
        hash2 = vaex.dataset.hash_array_data(self.indices[:])
        fp = vaex.cache.fingerprint(hash1, hash2, self.masked)
        return f'column-indexed-{fp}'

    @staticmethod
    def index(column, indices, direct_indices_map=None, masked=False):
        """Creates a new column indexed by indices which avoids nested indices

        :param df: Dataframe where column comes from
        :param column: Column object or numpy array
        :param name: name of column
        :param indices: ndarray with indices which rows to take
        :param direct_indices_map: cache of the nested indices (pass a dict for better performance), defaults to None
        :return: A column object which avoids two levels of indirection
        :rtype: ColumnIndexed
        """
        direct_indices_map = direct_indices_map if direct_indices_map is not None else {}
        if isinstance(column, ColumnIndexed):
            if id(column.indices) not in direct_indices_map:
                direct_indices = column.indices[indices]
                if masked:
                    # restored sentinel mask values
                    direct_indices[indices == -1] = -1
                direct_indices_map[id(column.indices)] = direct_indices
            else:
                direct_indices = direct_indices_map[id(column.indices)]
            return ColumnIndexed(column.column, direct_indices, masked=masked or column.masked)
        else:
            return ColumnIndexed(column, indices, masked=masked)

    def __len__(self):
        return len(self.indices)

    def trim(self, i1, i2):
        return ColumnIndexed(self.column, self.indices[i1:i2], masked=self.masked)

    def __arrow_array__(self, type=None):
        # TODO: without a copy we get a buserror
        # TODO2: weird, this path is not triggered anymore
        # values = self[:]
        # if hasattr(values, "to_arrow"):
        #     return values.to_arrow()
        # else:
        return pa.array(self)

    def to_numpy(self):
        return np.array(self[:])

    def __getitem__(self, slice):
        start, stop, step = slice.start, slice.stop, slice.step
        start = start or 0
        stop = stop or len(self)
        assert step in [None, 1]
        indices = self.indices[start:stop]
        ar_unfiltered = self.column
        if self.masked:
            mask = indices == -1
        if isinstance(ar_unfiltered, Column):
            # TODO: this is a workaround, since we do not require yet
            # that Column classes know how to deal with indices, we get
            # the minimal slice, and get those (not the most efficient)
            if self.masked:
                unmasked_indices = indices[~mask]
                if len(unmasked_indices) > 0:
                    i1, i2 = np.min(unmasked_indices), np.max(unmasked_indices)
                else:
                    i1, i2 = 0, 0
            else:
                i1, i2 = np.min(indices), np.max(indices)
            ar_unfiltered = ar_unfiltered[i1:i2+1]
            if self.masked:
                indices = indices - i1
            else:
                indices = indices - i1
        take_indices = indices
        if self.masked:
            # arrow and numpy do not like the negative indices, so we set them to 0
            take_indices = indices.copy()
            take_indices[mask] = 0
        if isinstance(ar_unfiltered, supported_arrow_array_types):
            ar = ar_unfiltered.take(vaex.array_types.to_arrow(take_indices))
        else:
            ar = ar_unfiltered[take_indices]
        assert not np.ma.isMaskedArray(indices)
        if self.masked:
            # TODO: we probably want to keep this as arrow array if it originally was
            return np.ma.array(ar, mask=mask)
        else:
            return ar


class ColumnConcatenatedLazy(Column):
    def __init__(self, expressions, dtype=None, shape=None, fill_value=None):
        self.is_masked = any([e.is_masked for e in expressions])
        self.fill_value = fill_value
        if self.is_masked and fill_value is None:
            for expression in expressions:
                if expression.is_masked:
                    try:
                        # fast path
                        self.fill_value = expression[0:1].fill_value
                        break
                    except:  # noqa
                        # slower path (we have to evaluate everything)
                        self.fill_value = expression.values.fill_value
                        break
            else:
                raise ValueError('Concatenating expressions with masked values, but no fill value is found')
        if dtype is None:
            dtypes = [e.dtype for e in expressions]

            any_strings = any([is_string_type(dtype) for dtype in dtypes])
            if any_strings:
                self.dtype = pa.string()  # TODO: how do we know it should not be large_string?
            else:
                # np.datetime64/timedelta64 and find_common_type don't mix very well
                if all([dtype == 'datetime64' for dtype in dtypes]):
                    self.dtype = dtypes[0]
                elif all([dtype == 'timedelta64' for dtype in dtypes]):
                    self.dtype = dtypes[0]
                else:
                    if all([dtype == dtypes[0] for dtype in dtypes]):  # find common types doesn't always behave well
                        self.dtype = dtypes[0]
                    if  any([dtype.kind in 'SU' for dtype in dtypes]):  # strings are also done manually
                        if all([dtype.kind in 'SU' for dtype in dtypes]):
                            index = np.argmax([dtype.itemsize for dtype in dtypes])
                            self.dtype = dtypes[index]
                        else:
                            index = np.argmax([df.columns[self.column_name].astype('O').astype('U').dtype.itemsize for df in dfs])
                            self.dtype = dfs[index].columns[self.column_name].astype('O').astype('U').dtype
                    else:
                        self.dtype = np.find_common_type([k.numpy for k in dtypes], [])
                    logger.debug("common type for %r is %r", dtypes, self.dtype)
            # make sure all expression are the same type
            self.expressions = [e if vaex.array_types.same_type(e.dtype, self.dtype) else e.astype(self.dtype) for e in expressions]
        else:
            # if dtype is given, we assume every expression/column is the same dtype
            self.dtype = dtype
            self.expressions = expressions[:]
        if shape is not None:
            self.shape = (len(self),) + shape
        else:
            self.shape = (len(self), ) + self.expressions[0].evaluate(0, 1, array_type='numpy', parallel=False).shape[1:]
            for i in range(1, len(self.expressions)):
                expression = self.expressions[i]
                shape_i = (len(self), ) + expressions[i].evaluate(0, 1, array_type='numpy', parallel=False).shape[1:]
                if self.shape != shape_i:
                    raise ValueError("shape of of expression %s, array index 0, is %r and is incompatible with the shape of the same column of array index %d, %r" % (self.expressions[0], self.shape, i, shape_i))

    def _fingerprint(self):
        fp = vaex.cache.fingerprint([k.fingerprint() for k in self.expressions])
        return f'column-concat-{fp}'

    def to_arrow(self, type=None):
        values = [e.values for e in self.expressions]
        chunks = [value if isinstance(value, pa.Array) else pa.array(value, type=type) for value in values]
        types = [chunk.type for chunk in chunks]

        # upcast if mixed types
        if pa.string() in types and pa.large_string() in types:
            def _arrow_string_upcast(array):
                if array.type == pa.large_string():
                    return array
                if array.type == pa.string():
                    import vaex.arrow.convert
                    column = vaex.arrow.convert.column_from_arrow_array(array)
                    column.indices = column.indices.astype(np.int64)
                    return pa.array(column)
                else:
                    raise ValueError('Not a string type: %r' % array)
            chunks = [_arrow_string_upcast(chunk) for chunk in chunks]

        return pa.chunked_array(chunks)

    def __len__(self):
        return sum(len(e.df) for e in self.expressions)

    def trim(self, i1, i2):
        start, stop = i1, i2
        i = 0  # the current expression
        offset = 0  # and the offset the current expression has wrt the dataframe
        # find the first expression we overlap with
        while start >= offset + len(self.expressions[i].df):
            offset += len(self.expressions[i].df)
            i += 1
        # do need more than 1?
        if stop > offset + len(self.expressions[i].df):
            # then add the first one
            expressions = [self.expressions[i][start-offset:]]
            offset += len(self.expressions[i].df)
            # we need more expression
            i += 1
            # keep adding complete expression till we find the it to be the last
            while stop > offset + len(self.expressions[i].df):
                offset += len(self.expressions[i].df)
                expressions.append(self.expressions[i])
                i += 1
            if stop > offset:  # add the tail part
                expressions.append(self.expressions[i][:stop-offset])
        else:
            # otherwise we only need a slice of the first
            expressions = [self.expressions[i][start-offset:stop-offset]]

        return ColumnConcatenatedLazy(expressions, self.dtype, self.shape[1:], fill_value=self.fill_value)

    def __arrow_array__(self, type=None):
        return pa.array(self[:], type=type)

    def __getitem__(self, slice):
        start, stop, step = slice.start, slice.stop, slice.step
        start = start or 0
        stop = stop or len(self)
        assert step in [None, 1]
        dtype = self.dtype
        if is_string_type(dtype):
            dtype = 'O'  # we store the strings in a dtype=object array
        expressions = iter(self.expressions)
        current_expression = next(expressions)
        offset = 0
        while start >= offset + len(current_expression.df):
            offset += len(current_expression.df)
            current_expression = next(expressions)
        # this is the fast path, no copy needed
        if stop <= offset + len(current_expression.df):
            if current_expression.df.filtered:  # TODO this may get slow! we're evaluating everything
                warnings.warn("might be slow, you have concatenated expressions with a filter set")
            values = current_expression.evaluate(i1=start - offset, i2=stop - offset, parallel=False)
        else:
            if self.is_masked:
                copy = np.ma.empty(stop - start, dtype=dtype)
                copy.fill_value = self.fill_value
            else:
                copy = np.zeros(stop - start, dtype=dtype)
            copy_offset = 0
            while offset < stop:  # > offset + len(current_expression):
                if current_expression.df.filtered:  # TODO this may get slow! we're evaluating everything
                    warnings.warn("might be slow, you have concatenated DataFrames with a filter set")
                part = current_expression.evaluate(i1=start-offset, i2=min(len(current_expression.df), stop - offset), parallel=False)
                copy[copy_offset:copy_offset + len(part)] = part
                offset += len(current_expression.df)
                copy_offset += len(part)
                start = offset
                if offset < stop:
                    current_expression = next(expressions)
            values = copy
        if is_string_type(dtype):
            return _to_string_column(values)
        else:
            return values


use_c_api = True


class ColumnString(Column):
    pass


class ColumnStringArray(Column):
    """Wraps a numpy array with dtype=object, containing all strings"""
    def __init__(self, array):
        # assert vaex.strings.all_strings(array, allow)
        self.array = array
        self.string_sequence = vaex.strings.StringArray(self.array)

    def __getitem__(self, slice):
        return ColumnString(self.array[slice])

    def to_numpy(self):
        return self.array

def _to_string(string_bytes):
    if six.PY2:
        return unicode(buffer(string_bytes))
    else:
        return bytes(memoryview(string_bytes)).decode('utf8')


def _is_stringy(x):
    if vaex.array_types.is_string(x):
        return True
    elif isinstance(x, ColumnString):
        return True
    elif isinstance(x, np.ndarray):
        if x.dtype.kind in 'US':
            return True
    return False


def _to_string_sequence(x, force=True):
    if isinstance(x, pa.DictionaryArray):
        x = x.dictionary.take(x.indices)  # equivalent to PyArrow 5.0.0's dictionary_decode() but backwards compatible
    if isinstance(x, pa.ChunkedArray):
        # turn into pa.Array, TODO: do we want this, this may result in a big mem copy
        table = pa.Table.from_arrays([x], ["single"])
        table_concat = table.combine_chunks()
        column = table_concat.columns[0]
        if column.num_chunks == 0:
            x = pa.array([], type=column.type)
        else:
            assert column.num_chunks == 1
            x = column.chunk(0)

    if isinstance(x, ColumnString):
        return x.string_sequence
    elif isinstance(x, pa.Array):
        from vaex.arrow import convert
        return convert.column_from_arrow_array(x).string_sequence
    elif isinstance(x, np.ndarray):
        mask = None
        if np.ma.isMaskedArray(x):
            mask = np.ma.getmaskarray(x)
            x = x.data
        if x.dtype == 'O':
            return vaex.strings.StringArray(x) if mask is None else vaex.strings.StringArray(x, mask)
        elif x.dtype.kind in 'US':
            x = x.astype('O')
            return vaex.strings.StringArray(x) if mask is None else vaex.strings.StringArray(x, mask)
        else:
            # This path is only required because the str_pandas wrapper uses NaN for missing values
            # see pandas_wrapper in functions.py
            if force:
                length = len(x)
                bytes = np.zeros((0,), dtype=np.uint8)
                indices = np.zeros((length+1,), dtype=np.int64)
                null_bitmap = np.ones(((length + 7) // 8,), dtype=np.uint8)
                return vaex.strings.StringList64(bytes, indices, length, 0, null_bitmap, 0)
            else:
                ValueError('unsupported dtype ' +str(x.dtype))
    elif isinstance(x, (vaex.superstrings.StringList32, vaex.superstrings.StringList64)):
        return x
    elif isinstance(x, (list, type)):
        return _to_string_sequence(np.array(x))
    else:
        raise ValueError('not a ColumnString or ndarray: ' + str(x))

def _to_string_column(x, force=True):
    ss = _to_string_sequence(x, force=force)
    if not isinstance(ss, (vaex.strings.StringList64, vaex.strings.StringList32)):
        ss = ss.to_arrow()
    return ColumnStringArrow.from_string_sequence(ss)

def _to_string_list_sequence(x):
    if isinstance(x, vaex.strings.StringListList):
        return x
    else:
        raise ValueError('not a StringListList')

def _asnumpy(ar):
    if isinstance(ar, np.ndarray):
        return ar
    else:
        return ar.to_numpy()


def _trim(column, i1, i2):
    if isinstance(column, np.ndarray):
        return column[i1:i2]
    else:
        return column.trim(i1, i2)

def _trim_bits(column, i1, i2):
    i1_bytes = i1 // 8
    offset = i1 - i1_bytes * 8
    i2_bytes = (i2 + 7) // 8
    if isinstance(column, np.ndarray):
        return column[i1_bytes:i2_bytes], offset
    else:
        return column.trim(i1_bytes, i2_bytes), offset

class ColumnStringArrow(ColumnString):
    """Column that unpacks the arrow string column on the fly"""
    def __init__(self, indices, bytes, length=None, offset=0, string_sequence=None, null_bitmap=None):
        self._string_sequence = string_sequence
        self.indices = indices
        self.offset = offset  # to avoid memory copies in trim
        self.bytes = bytes
        self.length = length if length is not None else len(indices) - 1
        if indices.dtype.kind == 'i' and indices.dtype.itemsize == 8:
            self.dtype = pa.large_string()
        elif indices.dtype.kind == 'i' and indices.dtype.itemsize == 4:
            self.dtype = pa.string()
        else:
            raise ValueError('unsupported index type' + str(indices.dtype))
        self.shape = (self.__len__(),)
        self.nbytes = self.bytes.nbytes + self.indices.nbytes
        self.null_bitmap = null_bitmap

        if not (self.indices.dtype.kind == 'i' and self.indices.dtype.itemsize in [4,8]):
            raise ValueError('unsupported index type' + str(self.indices.dtype))

    def _fingerprint(self):
        hash1 = vaex.dataset.hash_array_data(self.indices)
        hash2 = vaex.dataset.hash_array_data(self.bytes)
        fp = vaex.cache.fingerprint(hash1, hash2, self.length, self.offset, self.null_bitmap)
        return 'column-string-arrow-{fp}'

    def __arrow_array__(self, type=None):
        offsets = self.indices
        type = type or self.dtype
        if type == pa.string() and self.dtype == pa.large_string():
            offsets = offsets.astype(np.int32)  # downcast
        elif type == pa.large_string() and self.dtype == pa.string():
            type = pa.string()  # upcast

        # this code is very similar to vaex.arrow.convert.trim_buffers
        new_offset = self.offset % 8
        remove_offset = (self.offset // 8) * 8
        offsets = _asnumpy(_trim(offsets, remove_offset, self.offset + self.length + 1))
        first_offset = offsets[0]
        last_offset = offsets[new_offset + self.length]
        new_offsets = offsets - first_offset
        if self.null_bitmap is not None:
            null_bitmap, null_offset = _trim_bits(self.null_bitmap, self.offset, self.offset+self.length)
            assert null_offset == new_offset
            null_bitmap = pa.py_buffer(_asnumpy(null_bitmap))
        else:
            null_bitmap = None
        new_offsets_buffer = pa.py_buffer(new_offsets)
        bytes = pa.py_buffer(_asnumpy(_trim(self.bytes, first_offset, last_offset)))
        return pa.Array.from_buffers(type, self.length, [null_bitmap, new_offsets_buffer, bytes], offset=new_offset)

    @property
    def string_sequence(self):
        if self._string_sequence is None:
            if self.indices.dtype.kind == 'i' and self.indices.dtype.itemsize == 8:
                string_type = vaex.strings.StringList64
            elif self.indices.dtype.kind == 'i' and self.indices.dtype.itemsize == 4:
                string_type = vaex.strings.StringList32
            else:
                raise ValueError('unsupported index type' + str(self.indices.dtype))
            if self.offset == 0 and self.length == len(self.indices) - 1:
                # if the string is 'empty' with no indices
                # we just copy
                bytes = _asnumpy(self.bytes)
                indices = _asnumpy(self.indices)
                if self.null_bitmap is not None:
                    self._string_sequence = string_type(bytes, indices, self.length, self.offset, _asnumpy(self.null_bitmap), 0)
                else:
                    self._string_sequence = string_type(bytes, indices, self.length, self.offset)
            else:
                indices = _asnumpy(_trim(self.indices, self.offset, self.offset + self.length+1))
                if self.length:
                    byte_offset = indices[0]
                    byte_end = indices[self.length]
                    indices = indices - indices[0]
                    bytes = _asnumpy(_trim(self.bytes, byte_offset, byte_end))
                else:
                    bytes = _asnumpy(_trim(self.bytes, 0, 0))
                if self.null_bitmap is not None:
                    null_bitmap, null_offset = _trim_bits(self.null_bitmap, self.offset, self.offset+self.length)
                    self._string_sequence = string_type(bytes, indices, self.length, 0, _asnumpy(null_bitmap), null_offset)
                else:
                    self._string_sequence = string_type(bytes, indices, self.length, 0)
        return self._string_sequence

    def __len__(self):
        return self.length

    # TODO: in the future we might want to rely on these methods instead of special branches
    # in Expression for == and +. However, that gives different behaviour for dtype=object
    # (for instance you cannot use + when there is a None in it)
    def __eq__(self, other):
        if not isinstance(other, six.string_types):
            other = _to_string_sequence(other)
        return self.string_sequence.equals(other)

    def __add__(self, rhs):
        if not isinstance(rhs, six.string_types):
            rhs = _to_string_sequence(rhs)
        return self.from_string_sequence(self.string_sequence.concat(rhs))

    def __radd__(self, lhs):
        return self.from_string_sequence(self.string_sequence.concat_reverse(lhs))

    def __getitem__(self, slice):
        if isinstance(slice, int):
            return self.string_sequence.get(slice)
        elif isinstance(slice, np.ndarray):
            if np.ma.isMaskedArray(slice):
                if slice.dtype == np.bool_:
                    ss = self.string_sequence.index(slice.data & ~slice.mask)
                else:
                    ss = self.string_sequence.index(slice.data, slice.mask)
            else:
                ss = self.string_sequence.index(slice)
            return type(self).from_string_sequence(ss)
        else:
            start, stop, step = slice.start, slice.stop, slice.step
            start = start or 0
            stop = stop or len(self)
            assert step in [None, 1]
            return self.trim(start, stop)

    def to_numpy(self):
        return self.string_sequence.to_numpy()

    def trim(self, i1, i2):
        assert i2 >= i1
        return type(self)(self.indices, self.bytes, i2-i1, self.offset + i1, null_bitmap=self.null_bitmap)

    @classmethod
    def from_string_sequence(cls, string_sequence):
        s = string_sequence
        null_bitmap = s.null_bitmap
        if s.null_offset != 0:
            mask = ~s.mask()
            null_bitmap = np.frombuffer(pa.array(mask, pa.bool_()).buffers()[1], dtype='b')
        return cls(s.indices, s.bytes, s.length, s.offset, string_sequence=s, null_bitmap=null_bitmap)

    @classmethod
    def from_arrow(cls, ar):
        import vaex.arrow.convert
        return vaex.arrow.convert.column_from_arrow_array(ar)

    def _zeros_like(self):
        return ColumnStringArrow(np.zeros_like(self.indices), np.zeros_like(self.bytes), self.length, null_bitmap=self.null_bitmap)

    def get_mask(self):
        return self.string_sequence.mask()

    def astype(self, type):
        return self.to_numpy().astype(type)


import collections
import mmap
from typing import Any, Dict, List
import numpy as np

import pyarrow as pa
import pyarrow.csv
from dask.utils import parse_bytes
from frozendict import frozendict

import vaex.dataset
import vaex.settings
import vaex.file
from vaex.dataset import Dataset, DatasetFile
from .itertools import pmap, pwait, buffer, consume, filter_none
from .multithreading import get_main_io_pool


MB = 1024**2


def file_chunks(file, chunk_size, newline_readahead, fs=None, fs_options=None):
    """Bytes chunks, split by chunk_size bytes, on newline boundaries"""
    offset = 0
    # with open(file, 'rb') as file:\
    file_size = vaex.file.size(file, fs=fs, fs_options=fs_options)
    leftover = b''
    leftover_previous = b''

    with vaex.file.open(file, fs=fs, fs_options=fs_options) as file:
        begin_offset = 0
    
        done = False
        while not done:
            size_expected = chunk_size + newline_readahead
            data_extra = file.read(size_expected)
            if len(data_extra) == size_expected:
                sample = data_extra[chunk_size:]
                offset = vaex.superutils.find_byte(sample, ord(b'\n'))
                if offset == -1:
                    raise ValueError(f'Expected a newline within {newline_readahead} bytes, but not found, please increase newline_readahead')
                data, leftover = data_extra[:chunk_size+offset+1], data_extra[chunk_size+offset+1:]
            else:
                # read till the end
                data = data_extra
                done = True

            assert len(data) > 0
            def reader(data=data, leftover_previous=leftover_previous):
                chunk = leftover_previous + data
                return memoryview(chunk)
            leftover_previous = leftover
            yield reader


def file_chunks_mmap(file, chunk_size, newline_readahead):
    """Bytes chunks, split by chunk_size bytes, on newline boundaries
    
    Using memory mapping (which avoids a memcpy)
    """
    offset = 0
    with open(file, 'rb') as file:
        file.seek(0, 2)
        file_size = file.tell()
        file.seek(0)
        begin_offset = 0
        kwargs = {}
        if vaex.utils.osname == "windows":
            kwargs["access"] = mmap.ACCESS_READ
        else:
            kwargs["prot"] = mmap.PROT_READ

        file_map = mmap.mmap(file.fileno(), file_size, **kwargs)
        data = memoryview(file_map)
    
        done = False
        while not done:
            # find the next newline boundary
            end_offset = min(file_size, begin_offset + chunk_size)
            end_offset_readahead = min(file_size, begin_offset + chunk_size + newline_readahead)
            if end_offset_readahead < file_size:
                sample = data[end_offset:end_offset_readahead]
                offset = vaex.superutils.find_byte(sample, ord(b'\n'))
                if offset != -1:
                    end_offset += offset + 1 # include the newline
                else:
                    raise ValueError(f'Expected a newline within {newline_readahead} bytes, but not found, please increase newline_readahead')
            else:
                end_offset = file_size
            done = end_offset == file_size

            length = end_offset - begin_offset
            assert length > 0
            def reader(file_offset=begin_offset, length=length):
                return data[file_offset:file_offset+length]
            yield reader
            begin_offset = end_offset


def _row_count(chunk):
    ar = np.frombuffer(chunk, dtype=np.uint8)
    lines = vaex.superutils.count_byte(ar, ord(b'\n'))
    if ar[-1] != ord(b'\n'):
        lines += 1
    return lines


def _copy_or_create(cls, obj, **kwargs):
    if obj is not None:
        # take default from obj, buy kwargs have precendence
        kwargs = kwargs.copy()
        for name in dir(obj):
            value = getattr(obj, name)
            if not name.startswith('__') and not callable(value):
                if name not in kwargs:
                    kwargs[name] = value
    return cls(**kwargs)


def _get_kwargs(obj):
    kwargs = {}
    if obj is not None:
        # take default from obj, buy kwargs have precendence
        kwargs = kwargs.copy()
        for name in dir(obj):
            value = getattr(obj, name)
            if not name.startswith('__') and not callable(value):
                if name not in kwargs:
                    kwargs[name] = value
    return kwargs


@vaex.dataset.register
class DatasetCsvLazy(DatasetFile):
    snake_name = "arrow-csv-lazy"
    def __init__(self, path, chunk_size=10*MB, newline_readahead=1*MB, row_count=None, schema=None, read_options=None, parse_options=None, convert_options=None, schema_infer_fraction=0.001, fs=None, fs_options={}):
        super().__init__(path, fs=fs, fs_options=fs_options)
        try:
            codec = pa.Codec.detect(self.path)
        except Exception:
            codec = None
        if codec:
            raise NotImplementedError("We don't support compressed csv files for lazy reading, cannot read file: %s" % self.path)
        self._given_row_count = row_count
        self._row_count = None
        self._schema = schema
        self.chunk_size = parse_bytes(chunk_size)
        self.newline_readahead = parse_bytes(newline_readahead)


        self.read_options = read_options
        self.parse_options = parse_options
        self.convert_options = convert_options
        self.schema_infer_fraction = schema_infer_fraction
        self._infer_schema()

    @classmethod
    def quick_test(cls, path, fs_options={}, fs=None):
        path, options = vaex.file.split_options(path)
        return path.endswith('.csv')

    @classmethod
    def can_open(cls, path, fs_options={}, fs=None, group=None, **kwargs):
        return cls.quick_test(path, fs_options=fs_options, fs=fs)


    def _encode(self, encoding):
        spec = super()._encode(encoding)
        del spec["write"]
        return spec

    def __getstate__(self):
        state = super().__getstate__()
        state["read_options"] = self.read_options
        state["parse_options"] = self.parse_options
        state["convert_options"] = self.convert_options
        state["_given_row_count"] = self._given_row_count
        state["chunk_size"] = self.chunk_size
        state["newline_readahead"] = self.newline_readahead
        state["_row_count"] = self._row_count
        state["_schema"] = self._schema
        state["schema_infer_fraction"] = self.schema_infer_fraction
        return state

    def __setstate__(self, state):
        super().__setstate__(state)
        self._infer_schema()

    def leafs(self):
        return [self]

    @property
    def _fingerprint(self):
        paths = self.path
        fingerprint_file = vaex.file.fingerprint(self.path, fs_options=self.fs_options, fs=self.fs)
        # TODO: not sure if we need to put in the read_options
        fp = vaex.cache.fingerprint(fingerprint_file)
        return f'dataset-{self.snake_name}-{fp}'

    def _read_table(self, data, first, columns : List[str] = None):
        file_like = pa.input_stream(data)
        use_threads = True
        block_size = len(data)
        if first:
            read_options = _copy_or_create(pyarrow.csv.ReadOptions, self.read_options, use_threads=use_threads, block_size=block_size)
        else:
            read_options = _copy_or_create(pyarrow.csv.ReadOptions, self.read_options, use_threads=use_threads, block_size=block_size, column_names=self._column_names)

        if self._arrow_schema is None:
            convert_options = _copy_or_create(pyarrow.csv.ConvertOptions, self.convert_options, include_columns=columns)
        else:
            schema = pa.schema([(name, self._schema[name]) for name in columns])
            convert_options = _copy_or_create(pyarrow.csv.ConvertOptions, self.convert_options, column_types=schema, include_columns=columns)
        try:
            table = pyarrow.csv.read_csv(file_like, read_options=read_options, convert_options=convert_options)
        except pa.ArrowInvalid as e:
            import tempfile
            f = tempfile.NamedTemporaryFile(mode="wb", suffix=".csv")
            f.write(data)
            raise ValueError(
                f"Error reading csv file {self.path}, write offending chunk to: {f.name} (len={len(data)}, first={first}, columns={columns}, schema={schema}, encoding={read_options.encoding}, schema_infer_fraction={self.schema_infer_fraction}).\n"
                "Possible causes:\n"
                '  * This could be a file encoding error. Consider passing read_options=pyarrow.csv.ReadOptions(encoding="ISO-8859-1") or another encoding as argument.\n'
                "  * We might have inferred the wrong schema:\n"
                '     * Consider giving a schema hint by e.g. passing read_options=pyarrow.csv.ConvertOptions(column_types={"SomeId": pyarrow.string()}).\n'
                "     * Consider increasing schema_infer_fraction (e.g. schema_infer_fraction=1 to parse the whole file to infer the schema).\n"
            ) from e
        return table


    def _infer_schema(self):
        with vaex.file.open(self.path, fs_options=self.fs_options, fs=self.fs) as f:
            data = bytes(f.read(self.newline_readahead*3))
            offset = 0
            offset = data.find(b'\n', offset) + 1
            if offset == 0:
                raise ValueError("Cannot find newline in first %d bytes of file: %s" % (self.newline_readahead*3, self.path))
            offset = data.find(b'\n', offset) + 1
            if offset == 0:
                raise ValueError("Cannot find second newline in first %d bytes of file: %s" % (self.newline_readahead*3, self.path))
            f_first_two_lines = pa.input_stream(memoryview(data[:offset]))
            reader = pyarrow.csv.open_csv(f_first_two_lines, read_options=self.read_options, parse_options=self.parse_options, convert_options=self.convert_options)
            self._arrow_schema = reader.read_next_batch().schema
        self._column_names = list(self._arrow_schema.names)

        self._arrow_schema = None

        pool = get_main_io_pool()
        workers = pool._max_workers
        self._fragment_info = {}
        schemas: Dict[str, List[Any]] = collections.defaultdict(list)
        if self._schema is None and self._given_row_count is None and self._row_count is None:
            if self.fs is None and not self.fs_options:
                chunks = file_chunks_mmap(self.path, self.chunk_size, self.newline_readahead)
            else:
                chunks = file_chunks(self.path, self.chunk_size, self.newline_readahead, fs=self.fs, fs_options=self.fs_options)
            def process(i, chunk_reader):
                data = chunk_reader()
                if (i % int(1/self.schema_infer_fraction)) == 0:
                    table = self._read_table(data, first=i==0, columns=self._column_names)
                    for name, type in zip(table.schema.names, table.schema.types):
                        schemas[name].append(type)
                row_count = _row_count(data)
                if i == 0:
                    row_count -= 1  # we counted the header (TODO: depends on ReadOptions)
                self._fragment_info[i] = dict(row_count=row_count)
            consume(pwait(buffer(pmap(process, enumerate(chunks), pool=pool), workers+3)))
            row_start = 0
            for i in range(len(self._fragment_info)):
                row_end = row_start + self._fragment_info[i]['row_count']
                self._fragment_info[i]['row_start'] = row_start
                self._fragment_info[i]['row_end'] = row_end
                row_start = row_end
            self._row_count = row_start
            pool.map(process, enumerate(chunks))
        else:
            if self._row_count is None:
                self._row_count = self._given_row_count
        if self._schema is None:
            self._schema = {}
            from .schema import resolver_flexible

            for name, types in schemas.items():
                type, shape = resolver_flexible.resolve(types)
                if pa.types.is_null(type):
                    type = pa.int8()
                self._schema[name] = type
        self._arrow_schema = pa.schema([(name, type) for name, type in self._schema.items()])

        self._columns = {name: vaex.dataset.ColumnProxy(self, name, type) for name, type in self._schema.items()}

        self._ids = {}

    def hashed(self):
        raise NotImplementedError

    def slice(self, start, end):
        # TODO: we can be smarter here, and trim off some fragments
        if start == 0 and end == self.row_count:
            return self
        return vaex.dataset.DatasetSliced(self, start=start, end=end)

    def is_masked(self, column):
        return False

    def shape(self, column):
        return tuple()

    def __getitem__(self, item):
        if isinstance(item, slice):
            assert item.step in [1, None]
            return vaex.dataset.DatasetSliced(self, item.start or 0, item.stop or self.row_count)
        return self._columns[item]

    def close(self):
        # no need to close it, it seem
        pass

    def _chunk_producer(self, columns, chunk_size=None, reverse=False, start=0, end=None):
        pool = get_main_io_pool()
        
        first = True
        previous = None
        if self.fs is None and not self.fs_options:
            file_chunks_iterator = file_chunks_mmap(self.path, self.chunk_size, self.newline_readahead)
        else:
            file_chunks_iterator = file_chunks(self.path, self.chunk_size, self.newline_readahead, fs=self.fs, fs_options=self.fs_options)
        for i, reader in enumerate(file_chunks_iterator):
            fragment_info = self._fragment_info.get(i)
            # bail out/continue early
            if fragment_info:
                if start >= fragment_info['row_end']:  # we didn't find the beginning yet
                    continue
                # TODO, not triggered, should be <=
                if end < fragment_info['row_start']:  # we are past the end
                    # assert False
                    break

            def chunk_reader(reader=reader, first=first, previous=previous, fragment_info=fragment_info, i=i):
                bytes = reader()
                file_like = pa.input_stream(bytes)
                use_threads = True
                block_size = len(bytes)
                table = self._read_table(bytes, first=i==0, columns=columns)

                row_count = len(table)
                row_start = 0
                if i not in self._fragment_info:
                    if previous:
                        row_start, row_end, chunks = previous.result()
                        row_start = row_end
                    row_end = row_start + len(table)
                    self._fragment_info[i] = dict(
                            # begin_offset=begin_offset,
                            # end_offset=end_offset,
                            row_count=row_count,
                            row_start=row_start,
                            row_end=row_end,
                        )
                else:
                    row_start = self._fragment_info[i]['row_start']
                    row_end = self._fragment_info[i]['row_end']

                # this is the bail out when we didn't have the fragments info cached yet
                fragment_info = self._fragment_info[i]
                if start >= fragment_info['row_end']:  # we didn't find the beginning yet
                    return None
                if end <= fragment_info['row_start']:  # we are past the end
                    return None
                # print(start, end, fragment_info, row_start, row_end)


                if start > row_start:
                    # this means we have to cut off a piece of the beginning
                    if end < row_end:
                        # AND the end
                        length = end - row_start  # without the start cut off
                        length -= start - row_start  # correcting for the start cut off
                        # print(start, end, length, row_start, row_end)
                        table = table.slice(start - row_start, length)
                    else:
                        table = table.slice(start - row_start)
                else:
                    if end < row_end:
                        # we only need to cut off a piece of the end
                        length = end - row_start
                        table = table.slice(0, length)
                
                # table = table.combine_chunks()
                assert len(table)
                chunks = dict(zip(table.column_names, table.columns))
                return row_start, row_end, chunks


            previous = pool.submit(chunk_reader)
            yield previous
            first = False

    def chunk_iterator(self, columns, chunk_size=None, reverse=False, start=0, end=None):
        chunk_size = chunk_size or 1024*1024
        i1 = 0
        chunks_ready_list = []
        i1 = i2 = 0

        # althoug arrow uses threading, we still manage to get some extra performance out of it with more threads
        chunk_generator = self._chunk_producer(columns, chunk_size, start=start, end=end or self._row_count)
        for column in columns:
            if column not in self._columns:
                self._check_existence(column)
        for c1, c2, chunks in filter_none(pwait(buffer(chunk_generator, vaex.settings.main.thread_count_io//2+3))):
            chunks_ready_list.append(chunks)
            total_row_count = sum([len(list(k.values())[0]) for k in chunks_ready_list])
            if total_row_count > chunk_size:
                chunks_current_list, current_row_count = vaex.dataset._slice_of_chunks(chunks_ready_list, chunk_size)
                i2 += current_row_count
                yield i1, i2, vaex.dataset._concat_chunk_list(chunks_current_list)
                i1 = i2

        while chunks_ready_list:
            chunks_current_list, current_row_count = vaex.dataset._slice_of_chunks(chunks_ready_list, chunk_size)
            i2 += current_row_count
            yield i1, i2, vaex.dataset._concat_chunk_list(chunks_current_list)
            i1 = i2


@vaex.dataset.register
class DatasetCsv(DatasetFile):
    snake_name = "arrow-csv"

    def __init__(self, path, read_options=None, parse_options=None, convert_options=None, fs=None, fs_options={}):
        super(DatasetCsv, self).__init__(path, fs=fs, fs_options=fs_options)
        self.read_options = read_options
        self.parse_options = parse_options
        self.convert_options = convert_options
        self._read_file()

    @property
    def _fingerprint(self):
        fp = vaex.file.fingerprint(self.path, fs_options=self.fs_options, fs=self.fs)
        return f"dataset-{self.snake_name}-{fp}"

    def _read_file(self):
        import pyarrow.csv

        with vaex.file.open(self.path, fs=self.fs, fs_options=self.fs_options, for_arrow=True) as f:
            try:
                codec = pa.Codec.detect(self.path)
            except Exception:
                codec = None
            if codec:
                f = pa.CompressedInputStream(f, codec.name)
            self._arrow_table = pyarrow.csv.read_csv(f, read_options=self.read_options, parse_options=self.parse_options, convert_options=self.convert_options)
        self._columns = dict(zip(self._arrow_table.schema.names, self._arrow_table.columns))
        self._set_row_count()
        self._ids = frozendict({name: vaex.cache.fingerprint(self._fingerprint, name) for name in self._columns})

    def _encode(self, encoding):
        spec = super()._encode(encoding)
        del spec["write"]
        return spec

    def __getstate__(self):
        state = super().__getstate__()
        state["read_options"] = self.read_options
        state["parse_options"] = self.parse_options
        state["convert_options"] = self.convert_options
        return state

    def __setstate__(self, state):
        super().__setstate__(state)
        self._read_file()

    def close(self):
        pass


import dask.base

@dask.base.normalize_token.register(pyarrow.csv.ConvertOptions)
def _normalize(obj):
    kwargs = _get_kwargs(obj)
    return (pyarrow.csv.ConvertOptions, kwargs)

@dask.base.normalize_token.register(pyarrow.csv.ParseOptions)
def _normalize(obj):
    kwargs = _get_kwargs(obj)
    return (pyarrow.csv.ParseOptions, kwargs)

@dask.base.normalize_token.register(pyarrow.csv.ReadOptions)
def _normalize(obj):
    kwargs = _get_kwargs(obj)
    return (pyarrow.csv.ReadOptions, kwargs)



import logging
import numpy as np
import dask.base

import vaex.expression
from .utils import _split_and_combine_mask, as_flat_float

logger = logging.getLogger('vaex.selections')


def _select_replace(maskold, masknew):
    return masknew


def _select_and(maskold, masknew):
    return masknew if maskold is None else np.asarray(maskold) & np.asarray(masknew)


def _select_or(maskold, masknew):
    return masknew if maskold is None else np.asarray(maskold) | np.asarray(masknew)


def _select_xor(maskold, masknew):
    return masknew if maskold is None else np.asarray(maskold) ^ np.asarray(masknew)


def _select_subtract(maskold, masknew):
    return ~masknew if maskold is None else np.asarray(maskold) & ~np.asarray(masknew)


_select_functions = {"replace": _select_replace,
                     "and": _select_and,
                     "or": _select_or,
                     "xor": _select_xor,
                     "subtract": _select_subtract
                     }

class Selection(object):
    def __init__(self, previous_selection, mode):
        # we don't care about the previous selection if we simply replace the current selection
        self.previous_selection = previous_selection if mode != "replace" else None
        self.mode = mode

    def execute(self, datexecutor, execute_fully=False):
        if execute_fully and self.previous_selection:
            self.previous_selection.execute(executor=executor, execute_fully=execute_fully)

    def dependencies(self, df):
        deps = set()
        for e in self.expressions:
            deps |= df._selection_expression(e).dependencies()
        if self.previous_selection:
            deps |= self.previous_selection.dependencies(df)
        return deps

    def _depending_columns(self, ds):
        '''Find all columns that this selection depends on for df ds'''
        return self.dependencies(ds)


class SelectionDropNa(Selection):
    def __init__(self, drop_nan, drop_masked, expressions, previous_selection, mode):
        super(SelectionDropNa, self).__init__(previous_selection, mode)
        self.drop_nan = drop_nan
        self.drop_masked = drop_masked
        self.expressions = expressions

    def to_dict(self):
        previous = None
        if self.previous_selection:
            previous = self.previous_selection.to_dict()
        return dict(type="dropna", drop_nan=self.drop_nan, drop_masked=self.drop_masked, expressions=self.expressions,
                    mode=self.mode, previous_selection=previous)

    def _rename(self, df, old, new):
        pass  # TODO: do we need to rename the expressions?

    def evaluate(self, df, name, i1, i2, scope):
        if self.previous_selection:
            previous_mask = self.previous_selection.evaluate(df, name, i1, i2, scope)
        else:
            previous_mask = None
        mask = None
        for expressions in self.expressions:
            data = scope.evaluate(expressions)
            if mask is None:
                mask = np.full(len(data), True)
            if self.drop_nan and self.drop_masked:
                mask &= ~vaex.array_types.to_numpy(vaex.functions.isna(data))
            elif self.drop_nan:
                mask &= ~vaex.array_types.to_numpy(vaex.functions.isnan(data))
            elif self.drop_masked:
                mask &= ~vaex.array_types.to_numpy(vaex.functions.ismissing(data))
        if previous_mask is None:
            logger.debug("setting mask")
        else:
            logger.debug("combining previous mask with current mask using op %r", self.mode)
            mode_function = _select_functions[self.mode]
            mask = mode_function(previous_mask, mask)
        return mask


def _rename_expression_string(df, e, old, new):
    return vaex.expression.Expression(self.df, self.boolean_expression)._rename(old, new).expression

class SelectionExpression(Selection):
    def __init__(self,  boolean_expression, previous_selection, mode):
        super(SelectionExpression, self).__init__(previous_selection, mode)
        self.boolean_expression = str(boolean_expression)
        self.expressions = [self.boolean_expression]

    def _rename(self, df, old, new):
        boolean_expression = vaex.expression.Expression(df, self.boolean_expression)._rename(old, new).expression
        previous_selection = None
        if self.previous_selection:
            previous_selection = self.previous_selection._rename(df, old, new)
        return SelectionExpression(boolean_expression, previous_selection, self.mode)

    def to_dict(self):
        previous = None
        if self.previous_selection:
            previous = self.previous_selection.to_dict()
        return dict(type="expression", boolean_expression=str(self.boolean_expression), mode=self.mode, previous_selection=previous)

    def evaluate(self, df, name, i1, i2, scope):
        if self.previous_selection:
            previous_mask = self.previous_selection.evaluate(df, name, i1, i2, scope)
        else:
            previous_mask = None
        # result = df._evaluate_selection_mask(self.boolean_expression, i1, i2)
        result = scope.evaluate(self.boolean_expression)
        if isinstance(result, bool):
            N = i2 - i1
            current_mask = np.full(N, result)
        else:
            current_mask = vaex.array_types.to_numpy(result)
        if previous_mask is None:
            logger.debug("setting mask")
            mask = current_mask
        else:
            logger.debug("combining previous mask with current mask using op %r", self.mode)
            mode_function = _select_functions[self.mode]
            mask = mode_function(previous_mask, current_mask)
        return mask


class SelectionInvert(Selection):
    def __init__(self, previous_selection):
        super(SelectionInvert, self).__init__(previous_selection, "")
        self.expressions = []

    def to_dict(self):
        previous = None
        if self.previous_selection:
            previous = self.previous_selection.to_dict()
        return dict(type="invert", previous_selection=previous)

    def evaluate(self, df, name, i1, i2, scope):
        previous_mask = np.asarray(self.previous_selection.evaluate(df, name, i1, i2, scope))
        return ~previous_mask


class SelectionLasso(Selection):
    def __init__(self, boolean_expression_x, boolean_expression_y, xseq, yseq, previous_selection, mode):
        super(SelectionLasso, self).__init__(previous_selection, mode)
        self.boolean_expression_x = boolean_expression_x
        self.boolean_expression_y = boolean_expression_y
        self.xseq = xseq
        self.yseq = yseq
        self.expressions = [boolean_expression_x, boolean_expression_y]

    def evaluate(self, df, name, i1, i2, scope):
        if self.previous_selection:
            # previous_mask = df._evaluate_selection_mask(name, i1, i2, selection=self.previous_selection)
            previous_mask = self.previous_selection.evaluate(df, name, i1, i2, scope)
        else:
            previous_mask = None
        x, y = np.array(self.xseq, dtype=np.float64), np.array(self.yseq, dtype=np.float64)
        meanx = x.mean()
        meany = y.mean()
        radius = np.sqrt((meanx - x)**2 + (meany - y)**2).max()
        blockx = scope.evaluate(self.boolean_expression_x)
        blocky = scope.evaluate(self.boolean_expression_y)
        current_mask = np.full(len(blockx), False)
        # TODO: can we do without arrow->numpy conversion?
        blockx = vaex.array_types.to_numpy(blockx)
        blocky = vaex.array_types.to_numpy(blocky)
        (blockx, blocky), excluding_mask = _split_and_combine_mask([blockx, blocky])
        blockx = as_flat_float(blockx)
        blocky = as_flat_float(blocky)
        vaex.vaexfast.pnpoly(x, y, blockx, blocky, current_mask, meanx, meany, radius)
        if previous_mask is None:
            logger.debug("setting mask")
            mask = current_mask
        else:
            logger.debug("combining previous mask with current mask using op %r", self.mode)
            mode_function = _select_functions[self.mode]
            mask = mode_function(previous_mask, current_mask)
        if excluding_mask is not None:
            mask = mask & (~excluding_mask)
        return mask

    def to_dict(self):
        previous = None
        if self.previous_selection:
            previous = self.previous_selection.to_dict()
        return dict(type="lasso",
                    boolean_expression_x=str(self.boolean_expression_x),
                    boolean_expression_y=str(self.boolean_expression_y),
                    xseq=vaex.utils.make_list(self.xseq),
                    yseq=vaex.utils.make_list(self.yseq),
                    mode=self.mode,
                    previous_selection=previous)

def selection_from_dict(values):
    kwargs = dict(values)
    del kwargs["type"]
    if values["type"] == "lasso":
        kwargs["previous_selection"] = selection_from_dict(values["previous_selection"]) if values["previous_selection"] else None
        return SelectionLasso(**kwargs)
    elif values["type"] == "expression":
        kwargs["previous_selection"] = selection_from_dict(values["previous_selection"]) if values["previous_selection"] else None
        return SelectionExpression(**kwargs)
    elif values["type"] == "invert":
        kwargs["previous_selection"] = selection_from_dict(values["previous_selection"]) if values["previous_selection"] else None
        return SelectionInvert(**kwargs)
    elif values["type"] == "dropna":
        kwargs["previous_selection"] = selection_from_dict(values["previous_selection"]) if values["previous_selection"] else None
        return SelectionDropNa(**kwargs)
    else:
        raise ValueError("unknown type: %r, in dict: %r" % (values["type"], values))

@dask.base.normalize_token.register(Selection)
def normalize(e):
    return e.to_dict()



import numpy as np
import numbers
import six
import datetime
import pyarrow as pa
from vaex import datatype, struct

MAX_LENGTH = 50


def _trim_string(value):
    if len(value) > MAX_LENGTH:
        value = repr(value[:MAX_LENGTH-3])[:-1] + '...'
    return value

def _format_value(value):
    # print("value = ", value, type(value), isinstance(value, numbers.Number))
    if isinstance(value, pa.lib.Scalar):
        if datatype.DataType(value.type).is_struct:
            value = struct.format_struct_item_vaex_style(value)
        else:
            value = value.as_py()

        if value is None:
            return '--'
        else:
            return _trim_string(str(value))
    if isinstance(value, str):
        return _trim_string(str(value))
    elif isinstance(value, bytes):
        value = _trim_string(repr(value))
    elif isinstance(value, np.ma.core.MaskedConstant):
        return str(value)
    elif isinstance(value, np.datetime64):
        if np.isnat(value):
            value = 'NaT'
        else:
            value = ' '.join(str(value).split('T'))
        return value
    elif isinstance(value, np.timedelta64):
        if np.isnat(value):
            value = 'NaT'
        else:
            tmp = datetime.timedelta(seconds=value / np.timedelta64(1, 's'))
            ms = tmp.microseconds
            s = np.mod(tmp.seconds, 60)
            m = np.mod(tmp.seconds//60, 60)
            h = tmp.seconds // 3600
            d = tmp.days
            if ms:
                value = str('%i days %+02i:%02i:%02i.%i' % (d,h,m,s,ms))
            else:
                value = str('%i days %+02i:%02i:%02i' % (d,h,m,s))
        return value
    elif isinstance(value, numbers.Number):
        value = str(value)
    else:
        value = repr(value)
        value = _trim_string(value)
    return value
